from typing import Self

from pydantic import BaseModel

from infrahub.core.registry import registry
from infrahub.events import (
    GroupMemberAddedEvent,
    GroupMemberRemovedEvent,
    NodeCreatedEvent,
    NodeDeletedEvent,
    NodeUpdatedEvent,
)
from infrahub.trigger.models import (
    EventTrigger,
    ExecuteWorkflow,
    TriggerDefinition,
    TriggerType,
)
from infrahub.workflows.catalogue import (
    ACTION_ADD_NODE_TO_GROUP,
    ACTION_RUN_GENERATOR,
    ACTION_RUN_GENERATOR_GROUP_EVENT,
    REMOVE_ADD_NODE_FROM_GROUP,
)

from .constants import BranchScope, MemberAction, MemberUpdate, RelationshipMatch, ValueMatch


class EventGroupMember(BaseModel):
    id: str
    kind: str


class CoreAction(BaseModel):
    """CoreAction generic"""


class CoreGeneratorAction(CoreAction):
    generator_id: str


class CoreGroupAction(CoreAction):
    member_action: MemberAction
    group_id: str


class CoreTriggerRule(BaseModel):
    name: str
    branch_scope: BranchScope
    action: CoreAction
    active: bool


class CoreGroupTriggerRule(CoreTriggerRule):
    member_update: MemberUpdate
    group_id: str
    group_kind: str


class CoreNodeTriggerMatch(BaseModel):
    """Node Trigger Match Generic"""


class CoreNodeTriggerAttributeMatch(CoreNodeTriggerMatch):
    attribute_name: str
    value: str | None
    value_previous: str | None
    value_match: ValueMatch


class CoreNodeTriggerRelationshipMatch(CoreNodeTriggerMatch):
    relationship_name: str
    modification_type: RelationshipMatch
    peer: str | None


class CoreNodeTriggerRule(CoreTriggerRule):
    node_kind: str
    mutation_action: str
    matches: list[CoreNodeTriggerMatch]


class ActionTriggerRuleTriggerDefinition(TriggerDefinition):
    type: TriggerType = TriggerType.ACTION_TRIGGER_RULE

    @classmethod
    def from_trigger_rule(cls, trigger_rule: CoreTriggerRule) -> Self | None:
        if isinstance(trigger_rule, CoreNodeTriggerRule):
            return cls._from_node_trigger(trigger_rule=trigger_rule)

        if isinstance(trigger_rule, CoreGroupTriggerRule):
            return cls._from_group_trigger(trigger_rule=trigger_rule)

        return None

    @classmethod
    def _from_node_trigger(
        cls,
        trigger_rule: CoreNodeTriggerRule,
    ) -> Self:
        event_trigger = EventTrigger()

        match trigger_rule.mutation_action:
            case "created":
                event_trigger.events.add(NodeCreatedEvent.event_name)
            case "deleted":
                event_trigger.events.add(NodeDeletedEvent.event_name)
            case "updated":
                event_trigger.events.add(NodeUpdatedEvent.event_name)

        event_trigger.match = {"infrahub.node.kind": trigger_rule.node_kind}

        match trigger_rule.branch_scope:
            case BranchScope.DEFAULT_BRANCH:
                event_trigger.match["infrahub.branch.name"] = registry.default_branch
            case BranchScope.OTHER_BRANCHES:
                event_trigger.match["infrahub.branch.name"] = f"!{registry.default_branch}"

        related_matches: list[dict[str, str | list[str]]] = []
        for match in trigger_rule.matches:
            if isinstance(match, CoreNodeTriggerAttributeMatch):
                match_related: dict[str, str | list[str]] = {
                    "prefect.resource.role": "infrahub.node.attribute_update",
                    "infrahub.field.name": match.attribute_name,
                    "infrahub.attribute.action": ["added", "updated", "removed"],
                }

                match match.value_match:
                    case ValueMatch.VALUE:
                        match_related["infrahub.attribute.value"] = match.value or ""
                    case ValueMatch.VALUE_PREVIOUS:
                        match_related["infrahub.attribute.value_previous"] = match.value_previous or ""
                    case ValueMatch.VALUE_FULL:
                        match_related["infrahub.attribute.value"] = match.value or ""
                        match_related["infrahub.attribute.value_previous"] = match.value_previous or ""

            elif isinstance(match, CoreNodeTriggerRelationshipMatch):
                match_related = {
                    "prefect.resource.role": "infrahub.node.relationship_update",
                    "infrahub.field.name": match.relationship_name,
                }
                if isinstance(match.peer, str):
                    match_related["infrahub.relationship.peer_id"] = match.peer

                if match.modification_type != RelationshipMatch.UPDATED:
                    match_related["infrahub.relationship.peer_status"] = match.modification_type.value.name

            related_matches.append(match_related)

        event_trigger.match_related = related_matches or {}

        if isinstance(trigger_rule.action, CoreGeneratorAction):
            workflow = ExecuteWorkflow(
                workflow=ACTION_RUN_GENERATOR,
                parameters={
                    "branch_name": "{{ event.resource['infrahub.branch.name'] }}",
                    "generator_definition_id": trigger_rule.action.generator_id,
                    "node_ids": ["{{ event.resource['infrahub.node.id'] }}"],
                    "context": {
                        "__prefect_kind": "json",
                        "value": {
                            "__prefect_kind": "jinja",
                            "template": "{{ event.payload['context'] | tojson }}",
                        },
                    },
                },
            )
        elif isinstance(trigger_rule.action, CoreGroupAction):
            if trigger_rule.action.member_action == MemberAction.ADD_MEMBER:
                flow = ACTION_ADD_NODE_TO_GROUP
            else:
                flow = REMOVE_ADD_NODE_FROM_GROUP

            workflow = ExecuteWorkflow(
                workflow=flow,
                parameters={
                    "branch_name": "{{ event.resource['infrahub.branch.name'] }}",
                    "group_id": trigger_rule.action.group_id,
                    "node_id": "{{ event.resource['infrahub.node.id'] }}",
                    "context": {
                        "__prefect_kind": "json",
                        "value": {
                            "__prefect_kind": "jinja",
                            "template": "{{ event.payload['context'] | tojson }}",
                        },
                    },
                },
            )

        return cls(
            name=trigger_rule.name,
            trigger=event_trigger,
            actions=[workflow],
        )

    @classmethod
    def _from_group_trigger(
        cls,
        trigger_rule: CoreGroupTriggerRule,
    ) -> Self:
        event_trigger = EventTrigger()

        if trigger_rule.member_update == MemberUpdate.ADDED:
            event_trigger.events.add(GroupMemberAddedEvent.event_name)
        else:
            event_trigger.events.add(GroupMemberRemovedEvent.event_name)

        event_trigger.match = {"infrahub.node.kind": trigger_rule.group_kind, "infrahub.node.id": trigger_rule.group_id}

        match trigger_rule.branch_scope:
            case BranchScope.DEFAULT_BRANCH:
                event_trigger.match["infrahub.branch.name"] = registry.default_branch
            case BranchScope.OTHER_BRANCHES:
                event_trigger.match["infrahub.branch.name"] = f"!{registry.default_branch}"

        if isinstance(trigger_rule.action, CoreGeneratorAction):
            workflow = ExecuteWorkflow(
                workflow=ACTION_RUN_GENERATOR_GROUP_EVENT,
                parameters={
                    "branch_name": "{{ event.resource['infrahub.branch.name'] }}",
                    "generator_definition_id": trigger_rule.action.generator_id,
                    "members": {
                        "__prefect_kind": "json",
                        "value": {
                            "__prefect_kind": "jinja",
                            "template": "{{ event.payload['data']['members']| tojson }}",
                        },
                    },
                    "context": {
                        "__prefect_kind": "json",
                        "value": {
                            "__prefect_kind": "jinja",
                            "template": "{{ event.payload['context'] | tojson }}",
                        },
                    },
                },
            )

        return cls(
            name=trigger_rule.name,
            trigger=event_trigger,
            actions=[workflow],
        )



from __future__ import annotations

import re
from typing import TYPE_CHECKING

import ujson
from fastapi import APIRouter, Depends, Request
from lunr.index import Index
from pydantic import BaseModel

from infrahub import config
from infrahub.api.dependencies import get_current_user
from infrahub.config import (  # noqa: TC001
    AnalyticsSettings,
    ExperimentalFeaturesSettings,
    LoggingSettings,
    MainSettings,
)
from infrahub.core import registry
from infrahub.exceptions import NodeNotFoundError
from infrahub.workers.dependencies import get_installation_type

if TYPE_CHECKING:
    from infrahub.auth import AccountSession

router = APIRouter()


class ConfigAPI(BaseModel):
    main: MainSettings
    logging: LoggingSettings
    analytics: AnalyticsSettings
    experimental_features: ExperimentalFeaturesSettings
    sso: config.SSOInfo
    installation_type: str
    policy: config.PolicySettings


class InfoAPI(BaseModel):
    deployment_id: str
    version: str


@router.get("/config")
async def get_config() -> ConfigAPI:
    return ConfigAPI(
        main=config.SETTINGS.main,
        logging=config.SETTINGS.logging,
        analytics=config.SETTINGS.analytics,
        experimental_features=config.SETTINGS.experimental_features,
        sso=config.SETTINGS.security.public_sso_config,
        installation_type=get_installation_type(),
        policy=config.SETTINGS.policy,
    )


@router.get("/info")
async def get_info(request: Request, _: AccountSession = Depends(get_current_user)) -> InfoAPI:
    return InfoAPI(deployment_id=str(registry.id), version=request.app.version)


class SearchDocs:
    def __init__(self) -> None:
        self._title_documents: list[dict] = []
        self._heading_documents: list[dict] = []
        self._heading_index: Index | None = None

    def _load_json(self) -> None:
        """
        The structure of search-index.json is organized into an array of 3 arrays representing indexes for:
        [titleDocuments, headingDocuments, contentDocuments]

        For titleDocuments, it consists of an array of dictionaries with the following structure:
        {
            i: title_id,
            t: page_title,
            u: url,
            b: breadcrumb,
        }

        For headingDocuments, it is an array of dictionaries with the following structure:
        {
            i: incremental_id,
            t: section.title,
            u: url,
            h: section.hash,
            p: title_id,
        }

        For contentDocuments, it is an array of dictionaries with the following structure:
        {
            i: incremental_id,
            t: section.content,
            s: section.title or page_title,
            u: url,
            h: section.hash,
            p: title_id,
        }
        """

        try:
            with config.SETTINGS.main.docs_index_path.open(encoding="utf-8") as f:
                search_index = ujson.loads(f.read())
                self._title_documents = search_index[0]["documents"]
                heading_json = search_index[1]
                self._heading_documents = heading_json["documents"]
                self._heading_index = Index.load(heading_json["index"])
        except FileNotFoundError as exc:
            raise NodeNotFoundError(
                identifier=str(config.SETTINGS.main.docs_index_path),
                message="documentation index not found",
                node_type="file",
            ) from exc

    @property
    def heading_index(self) -> Index:
        if not self._heading_index:
            self._load_json()

        return self._heading_index

    @property
    def title_documents(self) -> list[dict]:
        if not self._title_documents:
            self._load_json()

        return self._title_documents

    @property
    def heading_documents(self) -> list[dict]:
        if not self._heading_documents:
            self._load_json()

        return self._heading_documents


search_docs_loader = SearchDocs()


def tokenize(text: str) -> list[str]:
    return re.findall(r"[^-\s]+", text.lower()) or []


def smart_queries(query: str) -> str:
    tokens = tokenize(query)
    if len(tokens) == 0:
        return ""

    term_required_and_wildcard = [f"{term}*" for term in tokens]

    return " ".join(term_required_and_wildcard)


class SearchResultAPI(BaseModel):
    title: str
    url: str
    breadcrumb: list[str]


@router.get("/search/docs", include_in_schema=False)
async def search_docs(
    query: str, limit: int | None = None, _: AccountSession = Depends(get_current_user)
) -> list[SearchResultAPI]:
    smart_query = smart_queries(query)
    search_results = search_docs_loader.heading_index.search(smart_query)
    heading_results = [
        next(doc for doc in search_docs_loader.heading_documents if doc["i"] == int(result["ref"]))
        for result in search_results
    ]

    if limit is not None:
        heading_results = heading_results[:limit]

    response_list: list[SearchResultAPI] = [
        SearchResultAPI(
            title=result["t"],
            url=result["u"] + result["h"],
            breadcrumb=next(
                doc["b"] + [doc["t"]] for doc in search_docs_loader.title_documents if doc["i"] == int(result["p"])
            ),
        )
        for result in heading_results
    ]

    return response_list



from typing import TYPE_CHECKING, Any

from infrahub.core.constants import NULL_VALUE, DiffAction, RelationshipCardinality
from infrahub.core.constants.database import DatabaseEdgeType
from infrahub.database import InfrahubDatabase
from infrahub.log import get_logger

from ..model.path import (
    CalculatedDiffs,
    EnrichedDiffProperty,
    EnrichedDiffRelationship,
    EnrichedDiffRoot,
    EnrichedDiffSingleRelationship,
)
from .interface import DiffEnricherInterface

if TYPE_CHECKING:
    from infrahub.core.schema import MainSchemaTypes

log = get_logger()


class DiffCardinalityOneEnricher(DiffEnricherInterface):
    """Clean up diffs for cardinality=one relationships to make them cleaner and more intuitive

    Final result is that each EnrichedDiffRelationship for a relationship of cardinality one
     - MUST have a single EnrichedDiffSingleRelationship (we'll call it the element)
     - the peer_id property of the element will be the latest non-null peer ID for this element
     - the element MUST have an EnrichedDiffProperty of property_type=IS_RELATED that correctly records
        the previous and new values of the peer ID for this element
     - changes to properties (IS_VISIBLE, etc) of a cardinality=one relationship are consolidated as well
    """

    def __init__(self, db: InfrahubDatabase):
        self.db = db
        self._node_schema_map: dict[str, MainSchemaTypes] = {}

    async def enrich(self, enriched_diff_root: EnrichedDiffRoot, calculated_diffs: CalculatedDiffs) -> None:  # noqa: ARG002
        self._node_schema_map = {}
        log.info("Beginning cardinality-one diff enrichment...")
        for diff_node in enriched_diff_root.nodes:
            for relationship_group in diff_node.relationships:
                if (
                    relationship_group.cardinality is RelationshipCardinality.ONE
                    and len(relationship_group.relationships) > 0
                ):
                    self.consolidate_cardinality_one_diff_elements(diff_relationship=relationship_group)
        log.info("Cardinality-one diff enrichment complete.")

    def _determine_action(self, previous_value: Any, new_value: Any) -> DiffAction:
        if previous_value == new_value:
            return DiffAction.UNCHANGED
        if previous_value in (None, NULL_VALUE):
            return DiffAction.ADDED
        if new_value in (None, NULL_VALUE):
            return DiffAction.REMOVED
        return DiffAction.UPDATED

    def _build_property_maps(
        self, diff_relationship: EnrichedDiffRelationship
    ) -> tuple[dict[DatabaseEdgeType, EnrichedDiffProperty], dict[DatabaseEdgeType, EnrichedDiffProperty]]:
        earliest_property_map: dict[DatabaseEdgeType, EnrichedDiffProperty] = {}
        latest_property_map: dict[DatabaseEdgeType, EnrichedDiffProperty] = {}
        for diff_rel_element in diff_relationship.relationships:
            for diff_property in diff_rel_element.properties:
                prop_type = diff_property.property_type
                current_earliest = earliest_property_map.get(prop_type)
                if not current_earliest:
                    earliest_property_map[prop_type] = diff_property
                elif diff_property.changed_at < current_earliest.changed_at:
                    earliest_property_map[prop_type] = diff_property
                # special handling for a REMOVE and ADD with the same timestamp to treat them as an update
                elif diff_property.changed_at == current_earliest.changed_at:
                    if diff_property.action is DiffAction.REMOVED and current_earliest.action is DiffAction.ADDED:
                        earliest_property_map[prop_type] = diff_property

                current_latest = latest_property_map.get(prop_type)
                if not current_latest:
                    latest_property_map[prop_type] = diff_property
                elif diff_property.changed_at > current_latest.changed_at:
                    latest_property_map[prop_type] = diff_property
                # special handling for a REMOVE and ADD with the same timestamp to treat them as an update
                elif diff_property.changed_at == current_latest.changed_at:
                    if diff_property.action is DiffAction.ADDED and current_latest.action is DiffAction.REMOVED:
                        latest_property_map[prop_type] = diff_property
        return (earliest_property_map, latest_property_map)

    def consolidate_cardinality_one_diff_elements(self, diff_relationship: EnrichedDiffRelationship) -> None:
        earliest_property_map, latest_property_map = self._build_property_maps(diff_relationship=diff_relationship)
        consolidated_properties = set()
        for prop_type, earliest_prop in earliest_property_map.items():
            latest_prop = latest_property_map[prop_type]
            # this means there was only one property of this type, so we keep it
            if earliest_prop is latest_prop:
                consolidated_properties.add(latest_prop)
                continue
            prop_action = self._determine_action(
                previous_value=earliest_prop.previous_value, new_value=latest_prop.new_value
            )
            consolidated_properties.add(
                EnrichedDiffProperty(
                    property_type=prop_type,
                    changed_at=latest_prop.changed_at,
                    previous_value=earliest_prop.previous_value,
                    new_value=latest_prop.new_value,
                    action=prop_action,
                    conflict=None,
                )
            )
        if consolidated_properties:
            element_actions = {element.action for element in diff_relationship.relationships}
            # check if this is a simultaneous update
            if len(diff_relationship.relationships) > 1 and {DiffAction.REMOVED, DiffAction.ADDED} <= element_actions:
                latest_element = [
                    element for element in diff_relationship.relationships if element.action is DiffAction.ADDED
                ][0]
                consolidated_element_action = DiffAction.UPDATED
            else:
                latest_element = max(diff_relationship.relationships, key=lambda elem: elem.changed_at)
                consolidated_element_action = latest_element.action
            diff_relationship.relationships = {
                EnrichedDiffSingleRelationship(
                    changed_at=latest_element.changed_at,
                    action=consolidated_element_action,
                    peer_id=latest_element.peer_id,
                    peer_label=latest_element.peer_label,
                    conflict=None,
                    properties=consolidated_properties,
                )
            }



from infrahub.core.validators.attribute.unique import AttributeUniquenessChecker
from infrahub.dependencies.interface import DependencyBuilder, DependencyBuilderContext


class SchemaAttributeUniqueConstraintDependency(DependencyBuilder[AttributeUniquenessChecker]):
    @classmethod
    def build(cls, context: DependencyBuilderContext) -> AttributeUniquenessChecker:
        return AttributeUniquenessChecker(db=context.db, branch=context.branch)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from infrahub.core.constants import InfrahubKind
from infrahub.core.query import QueryType

from .standard_node import StandardNodeQuery

if TYPE_CHECKING:
    from infrahub.core.task import Task
    from infrahub.database import InfrahubDatabase


class TaskNodeCreateQuery(StandardNodeQuery):
    name: str = "task_create"
    node: Task

    type: QueryType = QueryType.WRITE

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002
        node_type = self.node.get_type()
        self.params["node_prop"] = self.node.to_db()
        self.params["related_id"] = self.node.related.get_id()

        created_by = ""
        if self.node.account_id:
            query = """
            MATCH (a:%(account_kind)s {uuid: $account_id})
            """ % {"account_kind": InfrahubKind.ACCOUNT}
            self.add_to_query(query=query)
            self.params["account_id"] = self.node.account_id
            created_by = "CREATE (n)-[:CREATED_BY]->(a)"

        query = """
        MATCH (root:Root)
        MATCH (i:%(related_node_kind)s {uuid: $related_id})
        CREATE (n:%(node_type)s $node_prop)-[r:IS_PART_OF]->(root)
        CREATE (n)-[:IMPACTS]->(i)
        %(created_by)s
        """ % {"created_by": created_by, "node_type": node_type, "related_node_kind": self.node.related.get_kind()}
        self.add_to_query(query=query)

        self.return_labels = ["n"]


class TaskNodeQuery(StandardNodeQuery):
    name: str = "task_query"

    type: QueryType = QueryType.READ

    def __init__(self, ids: list[str], related_nodes: list[str], **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.ids = ids
        self.related_nodes = related_nodes
        self.params["related_nodes"] = self.related_nodes
        self.params["ids"] = self.ids

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002
        self.add_to_query(query="MATCH (n:Task)-[:IMPACTS]->(rn:Node)")
        if self.ids:
            self.add_to_query(query="WHERE n.uuid in $ids")
        if self.related_nodes:
            self.add_to_query(query="WHERE rn.uuid IN $related_nodes")

        self.order_by = ["n.created_at DESC"]
        self.return_labels = [
            "n",
            "rn",
        ]


class TaskNodeQueryWithLogs(TaskNodeQuery):
    name: str = "task_query_with_logs"

    type: QueryType = QueryType.READ

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002
        self.add_to_query(query="MATCH (n:Task)-[:IMPACTS]->(rn:Node)")
        if self.ids:
            self.add_to_query(query="WHERE n.uuid in $ids")
        if self.related_nodes:
            self.add_to_query(query="WHERE rn.uuid IN $related_nodes")

        query = """
        OPTIONAL MATCH (n)<-[r:RELATES_TO]-(l:TaskLog)
        WITH n, rn, COLLECT(l) as logs
        """

        self.add_to_query(query=query)
        self.order_by = ["n.created_at DESC"]

        self.return_labels = [
            "n",
            "logs",
            "rn",
        ]



from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING

from ..topological_sort import DependencyCycleExistsError, topological_sort
from .exceptions import SchemaImportError

if TYPE_CHECKING:
    from ..schema import NodeSchema


class InfrahubSchemaTopologicalSorter:
    def get_sorted_node_schema(
        self,
        schemas: Sequence[NodeSchema],
        required_relationships_only: bool = True,
        include: list[str] | None = None,
    ) -> list[set[str]]:
        relationship_graph: dict[str, set[str]] = {}
        for node_schema in schemas:
            if include and node_schema.kind not in include:
                continue
            relationship_graph[node_schema.kind] = set()
            for relationship_schema in node_schema.relationships:
                if required_relationships_only and relationship_schema.optional:
                    continue
                relationship_graph[node_schema.kind].add(relationship_schema.peer)

        try:
            return topological_sort(relationship_graph)
        except DependencyCycleExistsError as exc:
            raise SchemaImportError("Cannot import nodes. There are cycles in the dependency graph.") from exc



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Sequence

from infrahub.core.branch import Branch
from infrahub.core.branch.enums import BranchStatus
from infrahub.core.constants import (
    GLOBAL_BRANCH_NAME,
    BranchSupportType,
    InfrahubKind,
    RelationshipStatus,
    RepositoryInternalStatus,
)
from infrahub.core.migrations.shared import MigrationResult
from infrahub.core.query import Query, QueryType
from infrahub.core.timestamp import Timestamp

from ..query.attribute_add import AttributeAddQuery
from ..query.delete_element_in_schema import DeleteElementInSchemaQuery
from ..shared import GraphMigration

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase

default_branch = Branch(
    name="main",
    status=BranchStatus.OPEN,
    description="Default Branch",
    hierarchy_level=1,
    is_global=False,
    is_default=True,
    sync_with_git=False,
)


class Migration013ConvertCoreRepositoryWithCred(Query):
    name = "migration_013_convert_repository_with_cred"
    type = QueryType.WRITE

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        at = Timestamp()
        filters, params = at.get_query_filter_path()

        global_branch = Branch(
            name=GLOBAL_BRANCH_NAME,
            status=BranchStatus.OPEN,
            description="Global Branch",
            hierarchy_level=1,
            is_global=True,
            sync_with_git=False,
        )
        self.params.update(params)

        self.params["rel_props_new"] = {
            "branch": global_branch.name,
            "branch_level": global_branch.hierarchy_level,
            "status": RelationshipStatus.ACTIVE.value,
            "from": self.at.to_string(),
        }

        self.params["rel_props_del"] = {
            "branch": global_branch.name,
            "branch_level": global_branch.hierarchy_level,
            "status": RelationshipStatus.DELETED.value,
            "from": self.at.to_string(),
        }

        self.params["cred_node_props"] = {
            "branch_support": BranchSupportType.AGNOSTIC.value,
            "kind": InfrahubKind.PASSWORDCREDENTIAL,
            "namespace": "Core",
        }

        self.params["current_time"] = self.at.to_string()
        self.params["is_protected_default"] = False
        self.params["is_visible_default"] = True
        self.params["branch_support"] = BranchSupportType.AGNOSTIC.value

        self.params["rel_identifier"] = "gitrepository__credential"

        query = """
        // --------------------------------
        // Identify the git repositories to convert
        // --------------------------------
        MATCH path = (root:Root)<-[r:IS_PART_OF]-(node:%(git_repository)s)-[r9:HAS_ATTRIBUTE]-(a:Attribute)-[:HAS_VALUE]-(av:AttributeValue)
        WHERE a.name in ["username", "password"]
            AND av.value <> "NULL"
            AND all(r IN relationships(path) WHERE %(filters)s AND r.status = "active")
        WITH DISTINCT(node) as git_repo, root
        // --------------------------------
        // Prepare some nodes we'll need later
        // --------------------------------
        MERGE (is_protected_value:Boolean { value: $is_protected_default })
        MERGE (is_visible_value:Boolean { value: $is_visible_default })
        WITH git_repo, root, is_protected_value, is_visible_value
        // --------------------------------
        // Retrieve the name of the current repository
        // --------------------------------
        MATCH (git_repo)-[:HAS_ATTRIBUTE]-(git_attr_name:Attribute)-[:HAS_VALUE]->(git_name_value:AttributeValue)
        WHERE git_attr_name.name = "name"
        CALL (git_repo) {
            MATCH path1 = (git_repo)-[r1:HAS_ATTRIBUTE]-(git_attr_name2:Attribute)-[r2:HAS_VALUE]->(git_name_value2:AttributeValue)
            WHERE git_attr_name2.name = "name"
              AND all(r IN relationships(path1) WHERE %(filters)s)
            RETURN git_repo as n1, r1 as r11, r2 as r22, git_name_value2 as av1
            ORDER BY r1.branch_level DESC, r1.from DESC
            LIMIT 1
        }
        WITH  n1 as git_repo, r11 as r1, r22 as r2, av1 as git_name_value, root, is_protected_value, is_visible_value
        WHERE r1.status = "active" AND r2.status = "active"
        WITH DISTINCT(git_repo) as git_repo, root, is_protected_value, is_visible_value, git_name_value
        // --------------------------------
        // Create new CorePasswordCredential node
        // --------------------------------
        CREATE (cred:Node:%(credential)s:%(password_credential)s $cred_node_props )-[:IS_PART_OF $rel_props_new ]->(root)
        %(cred_guid)s
        // attribute: name
        CREATE (attr_name:Attribute { name: "name", branch_support: $branch_support })
        CREATE (attr_name)<-[:HAS_ATTRIBUTE $rel_props_new ]-(cred)
        CREATE (attr_name)-[:HAS_VALUE $rel_props_new ]->(git_name_value)
        CREATE (attr_name)-[:IS_PROTECTED $rel_props_new]->(is_protected_value)
        CREATE (attr_name)-[:IS_VISIBLE $rel_props_new]->(is_visible_value)
        // attribute: label
        CREATE (attr_lbl:Attribute { name: "label", branch_support: $branch_support })
        CREATE (attr_lbl)<-[:HAS_ATTRIBUTE $rel_props_new ]-(cred)
        CREATE (attr_lbl)-[:HAS_VALUE $rel_props_new ]->(git_name_value)
        CREATE (attr_lbl)-[:IS_PROTECTED $rel_props_new]->(is_protected_value)
        CREATE (attr_lbl)-[:IS_VISIBLE $rel_props_new]->(is_visible_value)
        // attribute: description
        CREATE (attr_desc:Attribute { name: "description", branch_support: $branch_support })
        MERGE (av_desc:AttributeValue { value: "Credential for " + git_name_value.value, is_default: true })
        CREATE (attr_desc)<-[:HAS_ATTRIBUTE $rel_props_new ]-(cred)
        CREATE (attr_desc)-[:HAS_VALUE $rel_props_new ]->(av_desc)
        CREATE (attr_desc)-[:IS_PROTECTED $rel_props_new]->(is_protected_value)
        CREATE (attr_desc)-[:IS_VISIBLE $rel_props_new]->(is_visible_value)
        %(attr_name_guid)s
        %(attr_label_guid)s
        %(attr_desc_guid)s
        WITH git_repo, cred
        // --------------------------------
        // Move Username and Password to the new credential node
        // --------------------------------
        MATCH (git_repo)-[r1:HAS_ATTRIBUTE]-(git_attr:Attribute)
        WHERE git_attr.name IN ["username", "password"]
        CREATE (cred)-[:HAS_ATTRIBUTE $rel_props_new ]->(git_attr)
        CREATE (git_repo)-[:HAS_ATTRIBUTE $rel_props_del ]->(git_attr)
        SET r1.to = $current_time
        WITH DISTINCT(git_repo) as git_repo, cred
        // --------------------------------
        // Create a new relationship between git_repo and Cred
        // --------------------------------
        CREATE (rel:Relationship { name: $rel_identifier, branch_support: $branch_support })
        %(rel_guid)s
        CREATE (git_repo)-[:IS_RELATED $rel_props_new]->(rel)<-[:IS_RELATED $rel_props_new]-(cred)
        """ % {
            "filters": filters,
            "cred_guid": db.render_uuid_generation(node_label="cred", node_attr="uuid", index=1),
            "attr_name_guid": db.render_uuid_generation(node_label="attr_name", node_attr="uuid", index=2),
            "attr_label_guid": db.render_uuid_generation(node_label="attr_lbl", node_attr="uuid", index=3),
            "attr_desc_guid": db.render_uuid_generation(node_label="attr_desc", node_attr="uuid", index=4),
            "rel_guid": db.render_uuid_generation(node_label="rel", node_attr="uuid", index=5),
            "git_repository": InfrahubKind.GENERICREPOSITORY,
            "credential": InfrahubKind.CREDENTIAL,
            "password_credential": InfrahubKind.PASSWORDCREDENTIAL,
        }
        self.add_to_query(query)
        self.return_labels = ["git_repo", "cred"]


class Migration013ConvertCoreRepositoryWithoutCred(Query):
    name = "migration_013_convert_repository_without_cred"
    type = QueryType.WRITE

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        at = Timestamp()
        filters, params = at.get_query_filter_path()

        global_branch = Branch(
            name=GLOBAL_BRANCH_NAME,
            status=BranchStatus.OPEN,
            description="Global Branch",
            hierarchy_level=1,
            is_global=True,
            sync_with_git=False,
        )
        self.params.update(params)

        self.params["rel_props_del"] = {
            "branch": global_branch.name,
            "branch_level": global_branch.hierarchy_level,
            "status": RelationshipStatus.DELETED.value,
            "from": self.at.to_string(),
        }

        self.params["current_time"] = self.at.to_string()

        query = """
        // --------------------------------
        // Identify the git repositories to convert
        // --------------------------------
        MATCH path = (node:%(git_repository)s)-[r9:HAS_ATTRIBUTE]-(a:Attribute)-[:HAS_VALUE]-(av:AttributeValue)
        WHERE a.name in ["username", "password"]
            AND av.value = "NULL"
            AND all(r IN relationships(path) WHERE %(filters)s AND r.status = "active")
        CALL (node) {
            MATCH (root:Root)<-[r:IS_PART_OF]-(node)
            WHERE %(filters)s
            RETURN node as n1, r as r1
            ORDER BY r.branch_level DESC, r.from DESC
            LIMIT 1
        }
        WITH n1 as node, r1 as rb
        WHERE rb.status = "active"
        WITH DISTINCT(node) as git_repo
        // --------------------------------
        // Delete Username and Password attributes
        // --------------------------------
        MATCH (git_repo)-[r1:HAS_ATTRIBUTE]-(git_attr:Attribute)
        WHERE git_attr.name IN ["username", "password"]
        CREATE (git_repo)-[:HAS_ATTRIBUTE $rel_props_del ]->(git_attr)
        SET r1.to = $current_time
        WITH DISTINCT(git_repo) as git_repo
        """ % {
            "filters": filters,
            "git_repository": InfrahubKind.GENERICREPOSITORY,
        }
        self.add_to_query(query)
        self.return_labels = ["git_repo"]


class Migration013DeleteUsernamePasswordGenericSchema(DeleteElementInSchemaQuery):
    name = "migration_013_delete_username_password_schema"
    type = QueryType.WRITE
    insert_return = False

    def __init__(self, **kwargs: Any):
        kwargs.pop("branch", None)

        super().__init__(
            element_names=["username", "password"],
            node_name="GenericRepository",
            node_namespace="Core",
            branch=default_branch,
            **kwargs,
        )


class Migration013DeleteUsernamePasswordReadWriteSchema(DeleteElementInSchemaQuery):
    name = "migration_013_delete_username_password_schema"
    type = QueryType.WRITE
    insert_return = False

    def __init__(self, **kwargs: Any):
        kwargs.pop("branch", None)

        super().__init__(
            element_names=["username", "password"],
            node_name="Repository",
            node_namespace="Core",
            branch=default_branch,
            **kwargs,
        )


class Migration013DeleteUsernamePasswordReadOnlySchema(DeleteElementInSchemaQuery):
    name = "migration_013_delete_username_password_schema"
    type = QueryType.WRITE
    insert_return = False

    def __init__(self, **kwargs: Any):
        kwargs.pop("branch", None)

        super().__init__(
            element_names=["username", "password"],
            node_name="ReadOnlyRepository",
            node_namespace="Core",
            branch=default_branch,
            **kwargs,
        )


class Migration013AddInternalStatusData(AttributeAddQuery):
    type = QueryType.WRITE

    def __init__(self, **kwargs: Any):
        kwargs.pop("branch", None)

        super().__init__(
            node_kind="CoreGenericRepository",
            attribute_name="internal_status",
            attribute_kind="Dropdown",
            branch_support=BranchSupportType.LOCAL.value,
            default_value=RepositoryInternalStatus.ACTIVE.value,
            branch=default_branch,
            **kwargs,
        )


class Migration013(GraphMigration):
    name: str = "013_convert_git_password_credential"
    queries: Sequence[type[Query]] = [
        Migration013ConvertCoreRepositoryWithCred,
        Migration013ConvertCoreRepositoryWithoutCred,
        Migration013DeleteUsernamePasswordGenericSchema,
        Migration013DeleteUsernamePasswordReadWriteSchema,
        Migration013DeleteUsernamePasswordReadOnlySchema,
        Migration013AddInternalStatusData,
    ]
    minimum_version: int = 12

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:  # noqa: ARG002
        result = MigrationResult()

        return result



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Mapping

from infrahub.core import registry
from infrahub.core.constants import RelationshipCardinality, RelationshipKind
from infrahub.core.constraint.node.runner import NodeConstraintRunner
from infrahub.core.manager import NodeManager
from infrahub.core.node import Node
from infrahub.core.protocols import CoreObjectTemplate
from infrahub.dependencies.registry import get_component_registry

if TYPE_CHECKING:
    from infrahub.core.branch import Branch
    from infrahub.core.relationship.model import RelationshipManager
    from infrahub.core.schema import MainSchemaTypes, NonGenericSchemaTypes, RelationshipSchema
    from infrahub.database import InfrahubDatabase


async def get_template_relationship_peers(
    db: InfrahubDatabase, template: CoreObjectTemplate, relationship: RelationshipSchema
) -> Mapping[str, CoreObjectTemplate]:
    """For a given relationship on the template, fetch the related peers."""
    template_relationship_manager: RelationshipManager = getattr(template, relationship.name)
    if relationship.cardinality == RelationshipCardinality.MANY:
        return await template_relationship_manager.get_peers(db=db, peer_type=CoreObjectTemplate)

    peers: dict[str, CoreObjectTemplate] = {}
    template_relationship_peer = await template_relationship_manager.get_peer(db=db, peer_type=CoreObjectTemplate)
    if template_relationship_peer:
        peers[template_relationship_peer.id] = template_relationship_peer
    return peers


async def extract_peer_data(
    db: InfrahubDatabase,
    template_peer: CoreObjectTemplate,
    obj_peer_schema: MainSchemaTypes,
    parent_obj: Node,
    current_template: CoreObjectTemplate,
) -> Mapping[str, Any]:
    obj_peer_data: dict[str, Any] = {}

    for attr in template_peer.get_schema().attribute_names:
        if attr not in obj_peer_schema.attribute_names:
            continue
        obj_peer_data[attr] = {"value": getattr(template_peer, attr).value, "source": template_peer.id}

    for rel in template_peer.get_schema().relationship_names:
        rel_manager: RelationshipManager = getattr(template_peer, rel)
        if (
            rel_manager.schema.kind not in [RelationshipKind.COMPONENT, RelationshipKind.PARENT]
            or rel_manager.schema.name not in obj_peer_schema.relationship_names
        ):
            continue

        if list(await rel_manager.get_peers(db=db)) == [current_template.id]:
            obj_peer_data[rel] = {"id": parent_obj.id}

    return obj_peer_data


async def handle_template_relationships(
    db: InfrahubDatabase,
    branch: Branch,
    obj: Node,
    template: CoreObjectTemplate,
    fields: list,
    constraint_runner: NodeConstraintRunner | None = None,
) -> None:
    if constraint_runner is None:
        component_registry = get_component_registry()
        constraint_runner = await component_registry.get_component(NodeConstraintRunner, db=db, branch=branch)

    for relationship in obj.get_relationships(kind=RelationshipKind.COMPONENT, exclude=fields):
        template_relationship_peers = await get_template_relationship_peers(
            db=db, template=template, relationship=relationship
        )
        if not template_relationship_peers:
            continue

        for template_relationship_peer in template_relationship_peers.values():
            # We retrieve peer schema for each peer in case we are processing a relationship which is based on a generic
            obj_peer_schema = registry.schema.get_node_schema(
                name=template_relationship_peer.get_schema().kind.removeprefix("Template"),
                branch=branch,
                duplicate=False,
            )
            obj_peer_data = await extract_peer_data(
                db=db,
                template_peer=template_relationship_peer,
                obj_peer_schema=obj_peer_schema,
                parent_obj=obj,
                current_template=template,
            )

            obj_peer = await Node.init(schema=obj_peer_schema, db=db, branch=branch)
            await obj_peer.new(db=db, **obj_peer_data)
            await constraint_runner.check(node=obj_peer, field_filters=list(obj_peer_data))
            await obj_peer.save(db=db)

            await handle_template_relationships(
                db=db,
                branch=branch,
                constraint_runner=constraint_runner,
                obj=obj_peer,
                template=template_relationship_peer,
                fields=fields,
            )


async def get_profile_ids(db: InfrahubDatabase, obj: Node) -> set[str]:
    if not hasattr(obj, "profiles"):
        return set()
    profile_rels = await obj.profiles.get_relationships(db=db)
    return {pr.peer_id for pr in profile_rels}


async def refresh_for_profile_update(
    db: InfrahubDatabase,
    branch: Branch,
    obj: Node,
    schema: NonGenericSchemaTypes,
    previous_profile_ids: set[str] | None = None,
) -> Node:
    if not hasattr(obj, "profiles"):
        return obj
    current_profile_ids = await get_profile_ids(db=db, obj=obj)
    if previous_profile_ids is None or previous_profile_ids != current_profile_ids:
        refreshed_node = await NodeManager.get_one_by_id_or_default_filter(
            db=db,
            kind=schema.kind,
            id=obj.get_id(),
            branch=branch,
            include_owner=True,
            include_source=True,
        )
        refreshed_node._node_changelog = obj.node_changelog
        return refreshed_node
    return obj


async def _do_create_node(
    node_class: type[Node],
    db: InfrahubDatabase,
    data: dict,
    schema: NonGenericSchemaTypes,
    fields_to_validate: list,
    branch: Branch,
    node_constraint_runner: NodeConstraintRunner,
) -> Node:
    obj = await node_class.init(db=db, schema=schema, branch=branch)
    await obj.new(db=db, **data)
    await node_constraint_runner.check(node=obj, field_filters=fields_to_validate)
    await obj.save(db=db)

    object_template = await obj.get_object_template(db=db)
    if object_template:
        await handle_template_relationships(
            db=db,
            branch=branch,
            template=object_template,
            obj=obj,
            fields=fields_to_validate,
        )
    return obj


async def create_node(
    data: dict,
    db: InfrahubDatabase,
    branch: Branch,
    schema: NonGenericSchemaTypes,
) -> Node:
    """Create a node in the database if constraint checks succeed."""

    component_registry = get_component_registry()
    node_constraint_runner = await component_registry.get_component(
        NodeConstraintRunner, db=db.start_session() if not db.is_transaction else db, branch=branch
    )
    node_class = Node
    if schema.kind in registry.node:
        node_class = registry.node[schema.kind]

    fields_to_validate = list(data)
    if db.is_transaction:
        obj = await _do_create_node(
            node_class=node_class,
            node_constraint_runner=node_constraint_runner,
            db=db,
            schema=schema,
            branch=branch,
            fields_to_validate=fields_to_validate,
            data=data,
        )
    else:
        async with db.start_transaction() as dbt:
            obj = await _do_create_node(
                node_class=node_class,
                node_constraint_runner=node_constraint_runner,
                db=dbt,
                schema=schema,
                branch=branch,
                fields_to_validate=fields_to_validate,
                data=data,
            )

    if await get_profile_ids(db=db, obj=obj):
        obj = await refresh_for_profile_update(db=db, branch=branch, schema=schema, obj=obj)

    return obj



from __future__ import annotations

import enum
from pathlib import Path
from typing import Protocol, runtime_checkable

import httpx
import ujson
from pydantic_settings import BaseSettings, SettingsConfigDict

from .utils import generate_request_filename


class RecorderType(str, enum.Enum):
    NONE = "none"
    JSON = "json"


@runtime_checkable
class Recorder(Protocol):
    def record(self, response: httpx.Response) -> None:
        """Record the response from Infrahub"""


class NoRecorder:
    @staticmethod
    def record(response: httpx.Response) -> None:
        """The NoRecorder just silently returns"""

    @classmethod
    def default(cls) -> NoRecorder:
        return cls()

    def __eq__(self, other: object) -> bool:
        return isinstance(other, NoRecorder)


class JSONRecorder(BaseSettings):
    model_config = SettingsConfigDict(env_prefix="INFRAHUB_JSON_RECORDER_")
    directory: str = "."
    host: str = ""

    def record(self, response: httpx.Response) -> None:
        self._set_url_host(response)
        filename = generate_request_filename(response.request)
        data = {
            "status_code": response.status_code,
            "method": response.request.method,
            "url": str(response.request.url),
            "headers": dict(response.request.headers),
            "response_content": response.content.decode("utf-8"),
            "request_content": response.request.content.decode("utf-8"),
        }

        with Path(f"{self.directory}/{filename}.json").open(mode="w", encoding="utf-8") as fobj:
            ujson.dump(data, fobj, indent=4, sort_keys=True)

    def _set_url_host(self, response: httpx.Response) -> None:
        if not self.host:
            return
        original = str(response.request.url)
        if response.request.url.port:
            modified = original.replace(
                f"{response.request.url.scheme}://{response.request.url.host}:",
                f"{response.request.url.scheme}://{self.host}:",
            )
        else:
            modified = original.replace(
                f"{response.request.url.scheme}://{response.request.url.host}/",
                f"{response.request.url.scheme}://{self.host}/",
            )

        response.request.url = httpx.URL(url=modified)



from __future__ import annotations

from typing import TYPE_CHECKING

from fastapi import APIRouter, Depends, Request
from starlette.responses import PlainTextResponse

from infrahub.api.dependencies import BranchParams, get_branch_params, get_current_user, get_db
from infrahub.core.constants import InfrahubKind
from infrahub.core.manager import NodeManager
from infrahub.database import InfrahubDatabase  # noqa: TC001
from infrahub.exceptions import CommitNotFoundError, PropagatedFromWorkerError
from infrahub.message_bus.messages import GitFileGet, GitFileGetResponse

if TYPE_CHECKING:
    from infrahub.core.protocols import CoreReadOnlyRepository, CoreRepository
    from infrahub.services import InfrahubServices


router = APIRouter(prefix="/file")


@router.get("/{repository_id:str}/{file_path:path}", response_class=PlainTextResponse)
async def get_file(
    request: Request,
    repository_id: str,
    file_path: str,
    branch_params: BranchParams = Depends(get_branch_params),
    db: InfrahubDatabase = Depends(get_db),
    commit: str | None = None,
    _: str = Depends(get_current_user),
) -> PlainTextResponse:
    """Retrieve a file from a git repository."""
    service: InfrahubServices = request.app.state.service

    repo: CoreRepository | CoreReadOnlyRepository = await NodeManager.get_one_by_id_or_default_filter(
        db=db,
        id=repository_id,
        kind=InfrahubKind.GENERICREPOSITORY,
        branch=branch_params.branch,
        at=branch_params.at,
    )

    commit = commit or repo.commit.value

    if not commit:
        raise CommitNotFoundError(identifier=repository_id, commit="", message="No commits found on this repository")

    message = GitFileGet(
        repository_id=repo.id,
        repository_name=str(repo.name.value),
        repository_kind=repo.get_kind(),
        commit=str(commit),
        file=file_path,
    )

    response = await service.message_bus.rpc(message=message, response_class=GitFileGetResponse)
    if response.data.http_code is not None:
        assert response.data.error_message is not None
        raise PropagatedFromWorkerError(message=response.data.error_message, http_code=response.data.http_code)

    return PlainTextResponse(content=response.data.content)



from rich.console import Console

from ..schema.repository import InfrahubRepositoryConfig


def list_transforms(config: InfrahubRepositoryConfig) -> None:
    console = Console()
    console.print(f"Python transforms defined in repository: {len(config.python_transforms)}")

    for transform in config.python_transforms:
        console.print(f"{transform.name} ({transform.file_path}::{transform.class_name})")



from __future__ import annotations

import hashlib
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any

from graphene import InputObjectType, Mutation
from graphene.types.mutation import MutationOptions
from typing_extensions import Self

from infrahub import config, lock
from infrahub.core.constants import InfrahubKind, MutationAction
from infrahub.core.constraint.node.runner import NodeConstraintRunner
from infrahub.core.manager import NodeManager
from infrahub.core.node.create import (
    create_node,
    get_profile_ids,
    refresh_for_profile_update,
)
from infrahub.core.schema import MainSchemaTypes, NodeSchema
from infrahub.core.schema.generic_schema import GenericSchema
from infrahub.core.schema.profile_schema import ProfileSchema
from infrahub.core.schema.template_schema import TemplateSchema
from infrahub.core.timestamp import Timestamp
from infrahub.database import retry_db_transaction
from infrahub.dependencies.registry import get_component_registry
from infrahub.events.generator import generate_node_mutation_events
from infrahub.exceptions import HFIDViolatedError, InitializationError, NodeNotFoundError
from infrahub.graphql.context import apply_external_context
from infrahub.graphql.field_extractor import extract_graphql_fields
from infrahub.lock import InfrahubMultiLock, build_object_lock_name
from infrahub.log import get_log_data, get_logger

from .node_getter.by_default_filter import MutationNodeGetterByDefaultFilter

if TYPE_CHECKING:
    from graphql import GraphQLResolveInfo

    from infrahub.core.branch import Branch
    from infrahub.core.node import Node
    from infrahub.core.schema.schema_branch import SchemaBranch
    from infrahub.database import InfrahubDatabase
    from infrahub.graphql.types.context import ContextInput

    from ..initialization import GraphqlContext


log = get_logger()

KINDS_CONCURRENT_MUTATIONS_NOT_ALLOWED = [InfrahubKind.GENERICGROUP]


@dataclass
class DeleteResult:
    node: Node
    mutation: InfrahubMutationMixin
    deleted_nodes: list[Node] = field(default_factory=list)


# ------------------------------------------
# Infrahub GraphQLType
# ------------------------------------------
class InfrahubMutationOptions(MutationOptions):
    schema: MainSchemaTypes | None = None

    @property
    def active_schema(self) -> MainSchemaTypes:
        if self.schema:
            return self.schema
        raise InitializationError("This class is not initialized with a schema")


class InfrahubMutationMixin:
    _meta: InfrahubMutationOptions

    @classmethod
    async def mutate(
        cls,
        root: dict,  # noqa: ARG003
        info: GraphQLResolveInfo,
        data: InputObjectType,
        context: ContextInput | None = None,
        **kwargs: dict[str, Any],
    ) -> Self:
        graphql_context: GraphqlContext = info.context
        await apply_external_context(graphql_context=graphql_context, context_input=context)

        obj = None
        mutation = None
        action = MutationAction.UNDEFINED
        deleted_nodes: list[Node] = []

        if "Create" in cls.__name__:
            obj, mutation = await cls.mutate_create(info=info, branch=graphql_context.branch, data=data)
            action = MutationAction.CREATED
        elif "Update" in cls.__name__:
            obj, mutation = await cls.mutate_update(info=info, branch=graphql_context.branch, data=data, **kwargs)
            action = MutationAction.UPDATED
        elif "Upsert" in cls.__name__:
            node_manager = NodeManager()
            node_getter_default_filter = MutationNodeGetterByDefaultFilter(
                db=graphql_context.db, node_manager=node_manager
            )
            obj, mutation, created = await cls.mutate_upsert(
                info=info,
                branch=graphql_context.branch,
                data=data,
                node_getter_default_filter=node_getter_default_filter,
                **kwargs,
            )
            if created:
                action = MutationAction.CREATED
            else:
                action = MutationAction.UPDATED
        elif "Delete" in cls.__name__:
            delete_result = await cls.mutate_delete(info=info, branch=graphql_context.branch, data=data, **kwargs)
            obj = delete_result.node
            mutation = delete_result.mutation
            deleted_nodes = delete_result.deleted_nodes

            action = MutationAction.DELETED
        else:
            raise ValueError(
                f"Unexpected class Name: {cls.__name__}, should end with Create, Update, Upsert, or Delete"
            )

        # Reset the time of the query to guarantee that all resolvers executed after this point will account for the changes
        graphql_context.at = Timestamp()

        if config.SETTINGS.broker.enable and graphql_context.background and obj.node_changelog.has_changes:
            log_data = get_log_data()
            request_id = log_data.get("request_id", "")

            events = await generate_node_mutation_events(
                node=obj,
                deleted_nodes=deleted_nodes,
                db=graphql_context.db,
                branch=graphql_context.branch,
                context=graphql_context.get_context(),
                request_id=request_id,
                action=action,
            )

            for event in events:
                graphql_context.background.add_task(graphql_context.active_service.event.send, event)

        return mutation

    @classmethod
    async def _call_mutate_create_object(
        cls, data: InputObjectType, db: InfrahubDatabase, branch: Branch, override_data: dict[str, Any] | None = None
    ) -> Node:
        """
        Wrapper around mutate_create_object to potentially activate locking.
        """
        schema_branch = db.schema.get_schema_branch(name=branch.name)
        lock_names = _get_kind_lock_names_on_object_mutation(
            kind=cls._meta.active_schema.kind, branch=branch, schema_branch=schema_branch, data=data
        )
        if lock_names:
            async with InfrahubMultiLock(lock_registry=lock.registry, locks=lock_names):
                return await cls.mutate_create_object(data=data, db=db, branch=branch, override_data=override_data)

        return await cls.mutate_create_object(data=data, db=db, branch=branch, override_data=override_data)

    @classmethod
    async def mutate_create(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        database: InfrahubDatabase | None = None,
        override_data: dict[str, Any] | None = None,
    ) -> tuple[Node, Self]:
        graphql_context: GraphqlContext = info.context
        db = database or graphql_context.db
        obj = await cls._call_mutate_create_object(data=data, db=db, branch=branch, override_data=override_data)
        result = await cls.mutate_create_to_graphql(info=info, db=db, obj=obj)
        return obj, result

    @classmethod
    @retry_db_transaction(name="object_create")
    async def mutate_create_object(
        cls,
        data: InputObjectType,
        db: InfrahubDatabase,
        branch: Branch,
        override_data: dict[str, Any] | None = None,
    ) -> Node:
        schema = cls._meta.active_schema
        if isinstance(schema, GenericSchema):
            raise ValueError(f"Node of generic schema `{schema.name=}` can not be instantiated.")
        create_data = dict(data)
        create_data.update(override_data or {})
        return await create_node(
            data=create_data,
            db=db,
            branch=branch,
            schema=schema,
        )

    @classmethod
    async def mutate_create_to_graphql(cls, info: GraphQLResolveInfo, db: InfrahubDatabase, obj: Node) -> Self:
        fields = extract_graphql_fields(info=info)
        result: dict[str, Any] = {"ok": True}
        if "object" in fields:
            result["object"] = await obj.to_graphql(db=db, fields=fields.get("object", {}))
        return cls(**result)

    @classmethod
    async def _call_mutate_update(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        db: InfrahubDatabase,
        obj: Node,
        skip_uniqueness_check: bool = False,
    ) -> tuple[Node, Self]:
        """
        Wrapper around mutate_update to potentially activate locking and call it within a database transaction.
        """

        schema_branch = db.schema.get_schema_branch(name=branch.name)
        lock_names = _get_kind_lock_names_on_object_mutation(
            kind=cls._meta.active_schema.kind, branch=branch, schema_branch=schema_branch, data=data
        )

        if db.is_transaction:
            if lock_names:
                async with InfrahubMultiLock(lock_registry=lock.registry, locks=lock_names):
                    obj = await cls.mutate_update_object(
                        db=db, info=info, data=data, branch=branch, obj=obj, skip_uniqueness_check=skip_uniqueness_check
                    )
            else:
                obj = await cls.mutate_update_object(
                    db=db, info=info, data=data, branch=branch, obj=obj, skip_uniqueness_check=skip_uniqueness_check
                )
            result = await cls.mutate_update_to_graphql(db=db, info=info, obj=obj)
            return obj, result

        async with db.start_transaction() as dbt:
            if lock_names:
                async with InfrahubMultiLock(lock_registry=lock.registry, locks=lock_names):
                    obj = await cls.mutate_update_object(
                        db=dbt,
                        info=info,
                        data=data,
                        branch=branch,
                        obj=obj,
                        skip_uniqueness_check=skip_uniqueness_check,
                    )
            else:
                obj = await cls.mutate_update_object(
                    db=dbt, info=info, data=data, branch=branch, obj=obj, skip_uniqueness_check=skip_uniqueness_check
                )
            result = await cls.mutate_update_to_graphql(db=dbt, info=info, obj=obj)
            return obj, result

    @classmethod
    @retry_db_transaction(name="object_update")
    async def mutate_update(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        database: InfrahubDatabase | None = None,
        node: Node | None = None,
    ) -> tuple[Node, Self]:
        graphql_context: GraphqlContext = info.context
        db = database or graphql_context.db

        obj = node or await NodeManager.find_object(
            db=db, kind=cls._meta.active_schema.kind, id=data.get("id"), hfid=data.get("hfid"), branch=branch
        )
        obj, result = await cls._call_mutate_update(info=info, data=data, db=db, branch=branch, obj=obj)

        return obj, result

    @classmethod
    async def mutate_update_object(
        cls,
        db: InfrahubDatabase,
        info: GraphQLResolveInfo,  # noqa: ARG003
        data: InputObjectType,
        branch: Branch,
        obj: Node,
        skip_uniqueness_check: bool = False,
    ) -> Node:
        component_registry = get_component_registry()
        node_constraint_runner = await component_registry.get_component(NodeConstraintRunner, db=db, branch=branch)

        before_mutate_profile_ids = await get_profile_ids(db=db, obj=obj)
        await obj.from_graphql(db=db, data=data)
        fields_to_validate = list(data)
        await node_constraint_runner.check(
            node=obj, field_filters=fields_to_validate, skip_uniqueness_check=skip_uniqueness_check
        )

        fields = list(data.keys())
        for field_to_remove in ("id", "hfid"):
            if field_to_remove in fields:
                fields.remove(field_to_remove)

        await obj.save(db=db, fields=fields)

        obj = await refresh_for_profile_update(
            db=db,
            branch=branch,
            obj=obj,
            previous_profile_ids=before_mutate_profile_ids,
            schema=cls._meta.active_schema,
        )
        return obj

    @classmethod
    async def mutate_update_to_graphql(
        cls,
        db: InfrahubDatabase,
        info: GraphQLResolveInfo,
        obj: Node,
    ) -> Self:
        fields_object = extract_graphql_fields(info=info)
        fields_object = fields_object.get("object", {})
        result: dict[str, Any] = {"ok": True}
        if fields_object:
            result["object"] = await obj.to_graphql(db=db, fields=fields_object)
        return cls(**result)

    @classmethod
    @retry_db_transaction(name="object_upsert")
    async def mutate_upsert(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        node_getter_default_filter: MutationNodeGetterByDefaultFilter,
        database: InfrahubDatabase | None = None,
    ) -> tuple[Node, Self, bool]:
        """
        First, check whether payload contains data identifying the node, such as id, hfid, or relevant fields for
        default_filter. If not, we will try to create the node, but this creation might fail if payload contains
        hfid fields (not `hfid` field itself) that would match an existing node in the database. In that case,
        we would update the node without rerunning uniqueness constraint.
        """

        schema = cls._meta.active_schema
        schema_name = schema.kind

        graphql_context: GraphqlContext = info.context
        db = database or graphql_context.db
        dict_data = dict(data)
        node = None

        if "id" in dict_data:
            node = await NodeManager.get_one(
                db=db, id=dict_data["id"], kind=schema_name, branch=branch, raise_on_error=True
            )
            updated_obj, mutation = await cls._call_mutate_update(
                info=info,
                data=data,
                db=db,
                branch=branch,
                obj=node,
            )
            return updated_obj, mutation, False

        if not schema.human_friendly_id and schema.default_filter is not None:
            node = await node_getter_default_filter.get_node(node_schema=schema, data=data, branch=branch)

        if "hfid" in data:
            node = await NodeManager.get_one_by_hfid(db=db, hfid=dict_data["hfid"], kind=schema_name, branch=branch)

        if node is not None:
            updated_obj, mutation = await cls._call_mutate_update(
                info=info,
                data=data,
                db=db,
                branch=branch,
                obj=node,
            )
            return updated_obj, mutation, False

        try:
            # This is a hack to avoid sitatuions where a node has an attribute or relationship called "pop"
            # which would have overridden the `pop` method of the InputObjectType object and as such would have
            # caused an error when trying to call `data.pop("hfid", None)`.
            # TypeError: 'NoneType' object is not callable
            data._pop = dict.pop.__get__(data, dict)
            data._pop("hfid", None)  # `hfid` is invalid for creation.
            created_obj, mutation = await cls.mutate_create(info=info, data=data, branch=branch)
            return created_obj, mutation, True
        except HFIDViolatedError as exc:
            # Only the HFID constraint has been violated, it means the node exists and we can update without rerunning constraints
            if len(exc.matching_nodes_ids) > 1:
                raise RuntimeError(f"Multiple {schema_name} nodes have the same hfid") from exc
            node_id = list(exc.matching_nodes_ids)[0]

            try:
                node = await NodeManager.get_one(
                    db=db, id=node_id, kind=schema_name, branch=branch, raise_on_error=True
                )
            except NodeNotFoundError as exc:
                if branch.is_default:
                    raise
                raise NodeNotFoundError(
                    node_type=exc.node_type,
                    identifier=exc.identifier,
                    branch_name=branch.name,
                    message=(
                        f"Node {exc.identifier} / {exc.node_type} uses this human-friendly ID, but does not exist on"
                        f" this branch. Please rebase this branch to access {exc.identifier} / {exc.node_type}"
                    ),
                ) from exc
            updated_obj, mutation = await cls._call_mutate_update(
                info=info,
                data=data,
                db=db,
                branch=branch,
                obj=node,
                skip_uniqueness_check=True,
            )
            return updated_obj, mutation, False

    @classmethod
    @retry_db_transaction(name="object_delete")
    async def mutate_delete(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
    ) -> DeleteResult:
        graphql_context: GraphqlContext = info.context

        obj = await NodeManager.find_object(
            db=graphql_context.db,
            kind=cls._meta.active_schema.kind,
            id=data.get("id"),
            hfid=data.get("hfid"),
            branch=branch,
        )

        async with graphql_context.db.start_transaction() as db:
            deleted = await NodeManager.delete(db=db, branch=branch, nodes=[obj])

        deleted_str = ", ".join([f"{d.get_kind()}({d.get_id()})" for d in deleted])
        log.info(f"nodes deleted: {deleted_str}")

        ok = True

        return DeleteResult(node=obj, mutation=cls(ok=ok), deleted_nodes=deleted)


class InfrahubMutation(InfrahubMutationMixin, Mutation):
    @classmethod
    def __init_subclass_with_meta__(
        cls,
        schema: NodeSchema | GenericSchema | ProfileSchema | TemplateSchema | None = None,
        _meta: InfrahubMutationOptions | None = None,
        **options: dict[str, Any],
    ) -> None:
        # Make sure schema is a valid NodeSchema Node Class
        if not isinstance(schema, NodeSchema | GenericSchema | ProfileSchema | TemplateSchema):
            raise ValueError(f"You need to pass a valid NodeSchema in '{cls.__name__}.Meta', received '{schema}'")

        if not _meta:
            _meta = InfrahubMutationOptions(cls)

        _meta.schema = schema

        super().__init_subclass_with_meta__(_meta=_meta, **options)


def _get_kinds_to_lock_on_object_mutation(kind: str, schema_branch: SchemaBranch) -> list[str]:
    """
    Return kinds for which we want to lock during creating / updating an object of a given schema node.
    Lock should be performed on schema kind and its generics having a uniqueness_constraint defined.
    If a generic uniqueness constraint is the same as the node schema one,
    it means node schema overrided this constraint, in which case we only need to lock on the generic.
    """

    node_schema = schema_branch.get(name=kind)

    schema_uc = None
    kinds = []
    if node_schema.uniqueness_constraints:
        kinds.append(node_schema.kind)
        schema_uc = node_schema.uniqueness_constraints

    if node_schema.is_generic_schema:
        return kinds

    generics_kinds = node_schema.inherit_from

    node_schema_kind_removed = False
    for generic_kind in generics_kinds:
        generic_uc = schema_branch.get(name=generic_kind).uniqueness_constraints
        if generic_uc:
            kinds.append(generic_kind)
            if not node_schema_kind_removed and generic_uc == schema_uc:
                # Check whether we should remove original schema kind as it simply overrides uniqueness_constraint
                # of a generic
                kinds.pop(0)
                node_schema_kind_removed = True
    return kinds


def _should_kind_be_locked_on_any_branch(kind: str, schema_branch: SchemaBranch) -> bool:
    """
    Check whether kind or any kind generic is in KINDS_TO_LOCK_ON_ANY_BRANCH.
    """

    if kind in KINDS_CONCURRENT_MUTATIONS_NOT_ALLOWED:
        return True

    node_schema = schema_branch.get(name=kind)
    if node_schema.is_generic_schema:
        return False

    for generic_kind in node_schema.inherit_from:
        if generic_kind in KINDS_CONCURRENT_MUTATIONS_NOT_ALLOWED:
            return True
    return False


def _hash(value: str) -> str:
    # Do not use builtin `hash` for lock names as due to randomization results would differ between
    # different processes.
    return hashlib.sha256(value.encode()).hexdigest()


def _get_kind_lock_names_on_object_mutation(
    kind: str, branch: Branch, schema_branch: SchemaBranch, data: InputObjectType
) -> list[str]:
    """
    Return objects kind for which we want to avoid concurrent mutation (create/update). Except for some specific kinds,
    concurrent mutations are only allowed on non-main branch as objects validations will be performed at least when merging in main branch.
    """

    if not branch.is_default and not _should_kind_be_locked_on_any_branch(kind=kind, schema_branch=schema_branch):
        return []

    if kind == InfrahubKind.GRAPHQLQUERYGROUP:
        # Lock on name as well to improve performances
        try:
            name = data.name.value
            return [build_object_lock_name(kind + "." + _hash(name))]
        except AttributeError:
            # We might reach here if we are updating a CoreGraphQLQueryGroup without updating the name,
            # in which case we would not need to lock. This is not supposed to happen as current `update`
            # logic first fetches the node with its name.
            return []

    lock_kinds = _get_kinds_to_lock_on_object_mutation(kind, schema_branch)
    lock_names = [build_object_lock_name(kind) for kind in lock_kinds]
    return lock_names


def _get_data_fields(data: InputObjectType) -> list[str]:
    return [field for field in data.keys() if field not in ["id", "hfid"]]



from __future__ import annotations

import logging
import os
from typing import TYPE_CHECKING

import typer
from deepdiff import DeepDiff
from infrahub_sdk.async_typer import AsyncTyper
from prefect.client.orchestration import get_client
from rich import print as rprint

from infrahub import config
from infrahub.core.initialization import create_anonymous_role, create_default_account_groups, initialize_registry
from infrahub.core.manager import NodeManager
from infrahub.core.protocols import CoreAccount, CoreObjectPermission
from infrahub.menu.menu import default_menu
from infrahub.menu.models import MenuDict
from infrahub.menu.repository import MenuRepository
from infrahub.menu.utils import create_default_menu
from infrahub.trigger.tasks import trigger_configure_all
from infrahub.workflows.initialization import (
    setup_blocks,
    setup_deployments,
    setup_worker_pools,
)

from .db import initialize_internal_schema, migrate_database, update_core_schema

if TYPE_CHECKING:
    from infrahub.cli.context import CliContext
    from infrahub.database import InfrahubDatabase

app = AsyncTyper()


@app.command(name="upgrade")
async def upgrade_cmd(
    ctx: typer.Context,
    config_file: str = typer.Argument("infrahub.toml", envvar="INFRAHUB_CONFIG"),
    check: bool = typer.Option(False, help="Check the state of the system without upgrading."),
) -> None:
    """Upgrade Infrahub to the latest version."""

    logging.getLogger("infrahub").setLevel(logging.WARNING)
    logging.getLogger("neo4j").setLevel(logging.ERROR)
    logging.getLogger("prefect").setLevel(logging.ERROR)
    os.environ["PREFECT_SERVER_ANALYTICS_ENABLED"] = "false"

    config.load_and_exit(config_file_name=config_file)

    context: CliContext = ctx.obj
    dbdriver = await context.init_db(retry=1)

    await initialize_registry(db=dbdriver)

    # NOTE add step to validate if the database and the task manager are reachable

    # -------------------------------------------
    # Add pre-upgrade  validation
    # -------------------------------------------

    # -------------------------------------------
    # Upgrade Infrahub Database and Schema
    # -------------------------------------------

    if not await migrate_database(db=dbdriver, initialize=False, check=check):
        # A migration failed, stop the upgrade process
        rprint("Upgrade cancelled due to migration failure.")
        await dbdriver.close()
        return

    await initialize_internal_schema()
    await update_core_schema(db=dbdriver, initialize=False)

    # -------------------------------------------
    # Upgrade Internal Objects, generated and managed by Infrahub
    # -------------------------------------------
    await upgrade_menu(db=dbdriver)
    await upgrade_permissions(db=dbdriver)

    # -------------------------------------------
    # Upgrade External system : Task Manager
    # -------------------------------------------
    async with get_client(sync_client=False) as client:
        await setup_blocks()
        await setup_worker_pools(client=client)
        await setup_deployments(client=client)
        await trigger_configure_all()

    await dbdriver.close()


async def upgrade_menu(db: InfrahubDatabase) -> None:
    menu_repository = MenuRepository(db=db)
    menu_nodes = await menu_repository.get_menu_db()
    menu_items = await menu_repository.get_menu(nodes=menu_nodes)
    default_menu_dict = MenuDict.from_definition_list(default_menu)

    if not menu_nodes:
        await create_default_menu(db=db)
        return

    diff_menu = DeepDiff(menu_items.to_rest(), default_menu_dict.to_rest(), ignore_order=True)

    if not diff_menu:
        rprint("Menu Up to date, nothing to update")
        return

    await menu_repository.update_menu(existing_menu=menu_items, new_menu=default_menu_dict, menu_nodes=menu_nodes)
    rprint("Menu has been updated")


async def upgrade_permissions(db: InfrahubDatabase) -> None:
    existing_permissions = await NodeManager.query(schema=CoreObjectPermission, db=db, limit=1)
    if existing_permissions:
        rprint("Permissions Up to date, nothing to update")
        return

    await setup_permissions(db=db)
    rprint("Permissions have been updated")


async def setup_permissions(db: InfrahubDatabase) -> None:
    existing_accounts = await NodeManager.query(schema=CoreAccount, db=db, limit=1)
    await create_default_account_groups(db=db, admin_accounts=existing_accounts)

    if config.SETTINGS.main.allow_anonymous_access:
        await create_anonymous_role(db=db)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from pydantic import BaseModel

from infrahub.core.constants import BranchSupportType, RelationshipStatus
from infrahub.core.graph.schema import GraphNodeRelationships, GraphRelDirection
from infrahub.core.query import Query, QueryType

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase


class SchemaNodeInfo(BaseModel):
    name: str
    namespace: str
    branch_support: str = BranchSupportType.AWARE.value
    labels: list[str]
    kind: str


class NodeDuplicateQuery(Query):
    name = "node_duplicate"
    type = QueryType.WRITE
    insert_return: bool = False

    def __init__(
        self,
        previous_node: SchemaNodeInfo,
        new_node: SchemaNodeInfo,
        **kwargs: Any,
    ) -> None:
        self.previous_node = previous_node
        self.new_node = new_node

        super().__init__(**kwargs)

    def render_match(self) -> str:
        labels_str = ":".join(self.previous_node.labels)
        query = f"""
        // Find all the active nodes
        MATCH (node:{labels_str})
        WITH DISTINCT node
        """

        return query

    @staticmethod
    def _render_sub_query_per_rel_type(rel_name: str, rel_type: str, rel_dir: GraphRelDirection) -> str:
        subquery = [
            f"WITH peer_node, {rel_name}, active_node, new_node",
            f'WHERE type({rel_name}) = "{rel_type}"',
        ]
        if rel_dir in [GraphRelDirection.OUTBOUND, GraphRelDirection.EITHER]:
            subquery.append(f"""
                CREATE (new_node)-[new_active_edge:{rel_type} $rel_props_new ]->(peer_node)
                SET new_active_edge.branch = CASE WHEN {rel_name}.branch = "-global-" THEN "-global-" ELSE $branch END
                SET new_active_edge.branch_level = CASE WHEN {rel_name}.branch = "-global-" THEN {rel_name}.branch_level ELSE $branch_level END
                SET new_active_edge.hierarchy = COALESCE({rel_name}.hierarchy, NULL)
                """)
            subquery.append(f"""
                CREATE (active_node)-[deleted_edge:{rel_type} $rel_props_prev ]->(peer_node)
                SET deleted_edge.branch = CASE WHEN {rel_name}.branch = "-global-" THEN "-global-" ELSE $branch END
                SET deleted_edge.branch_level = CASE WHEN {rel_name}.branch = "-global-" THEN {rel_name}.branch_level ELSE $branch_level END
                SET deleted_edge.hierarchy = COALESCE({rel_name}.hierarchy, NULL)
                """)
        elif rel_dir in [GraphRelDirection.INBOUND, GraphRelDirection.EITHER]:
            subquery.append(f"""
                CREATE (new_node)<-[new_active_edge:{rel_type} $rel_props_new ]-(peer_node)
                SET new_active_edge.branch = CASE WHEN {rel_name}.branch = "-global-" THEN "-global-" ELSE $branch END
                SET new_active_edge.branch_level = CASE WHEN {rel_name}.branch = "-global-" THEN {rel_name}.branch_level ELSE $branch_level END
                SET new_active_edge.hierarchy = COALESCE({rel_name}.hierarchy, NULL)
                """)
            subquery.append(f"""
                CREATE (active_node)<-[deleted_edge:{rel_type} $rel_props_prev ]-(peer_node)
                SET deleted_edge.branch = CASE WHEN {rel_name}.branch = "-global-" THEN "-global-" ELSE $branch END
                SET deleted_edge.branch_level = CASE WHEN {rel_name}.branch = "-global-" THEN {rel_name}.branch_level ELSE $branch_level END
                SET deleted_edge.hierarchy = COALESCE({rel_name}.hierarchy, NULL)
                """)
        subquery.append("RETURN peer_node as p2")
        return "\n".join(subquery)

    @classmethod
    def _render_sub_query_out(cls) -> tuple[str, str]:
        rel_name = "rel_outband"
        sub_query_out_args = f"peer_node, {rel_name}, active_node, new_node"
        sub_queries_out = [
            cls._render_sub_query_per_rel_type(rel_name=rel_name, rel_type=rel_type, rel_dir=GraphRelDirection.OUTBOUND)
            for rel_type, field_info in GraphNodeRelationships.model_fields.items()
            if field_info.default.direction in (GraphRelDirection.OUTBOUND, GraphRelDirection.EITHER)
        ]
        sub_query_out = "\nUNION\n".join(sub_queries_out)
        return sub_query_out, sub_query_out_args

    @classmethod
    def _render_sub_query_in(cls) -> tuple[str, str]:
        rel_name = "rel_inband"
        sub_query_in_args = f"peer_node, {rel_name}, active_node, new_node"
        sub_queries_in = [
            cls._render_sub_query_per_rel_type(rel_name=rel_name, rel_type=rel_type, rel_dir=GraphRelDirection.INBOUND)
            for rel_type, field_info in GraphNodeRelationships.model_fields.items()
            if field_info.default.direction in (GraphRelDirection.INBOUND, GraphRelDirection.EITHER)
        ]
        sub_query_in = "\nUNION\n".join(sub_queries_in)
        return sub_query_in, sub_query_in_args

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        branch_filter, branch_params = self.branch.get_query_filter_path(at=self.at.to_string())
        self.params.update(branch_params)

        self.params["new_node"] = self.new_node.model_dump()
        self.params["previous_node"] = self.previous_node.model_dump()

        self.params["current_time"] = self.at.to_string()
        self.params["branch"] = self.branch.name
        self.params["branch_level"] = self.branch.hierarchy_level
        self.params["branch_support"] = self.new_node.branch_support

        self.params["rel_props_new"] = {
            "status": RelationshipStatus.ACTIVE.value,
            "from": self.at.to_string(),
        }

        self.params["rel_props_prev"] = {
            "status": RelationshipStatus.DELETED.value,
            "from": self.at.to_string(),
        }

        sub_query_out, sub_query_out_args = self._render_sub_query_out()
        sub_query_in, sub_query_in_args = self._render_sub_query_in()

        self.add_to_query(self.render_match())

        # ruff: noqa: E501
        query = """
        CALL (node) {
            MATCH (root:Root)<-[r:IS_PART_OF]-(node)
            WHERE %(branch_filter)s
            RETURN node as n1, r as r1
            ORDER BY r.branch_level DESC, r.from DESC
            LIMIT 1
        }
        WITH n1 as active_node, r1 as rb
        WHERE rb.status = "active"
        CREATE (new_node:Node:%(labels)s { uuid: active_node.uuid, kind: $new_node.kind, namespace: $new_node.namespace, branch_support: $new_node.branch_support })
        WITH active_node, new_node
        // Process Outbound Relationship
        MATCH (active_node)-[]->(peer)
        WITH DISTINCT active_node, new_node, peer
        CALL (active_node, peer) {
            MATCH (active_node)-[r]->(peer)
            WHERE %(branch_filter)s
            RETURN active_node as n1, r as rel_outband1, peer as p1
            ORDER BY r.branch_level DESC, r.from DESC
            LIMIT 1
        }
        WITH n1 as active_node, rel_outband1 as rel_outband, p1 as peer_node, new_node
        WHERE rel_outband.status = "active" AND rel_outband.to IS NULL
        CALL (%(sub_query_out_args)s) {
            %(sub_query_out)s
        }
        WITH p2 as peer_node, rel_outband, active_node, new_node
        FOREACH (i in CASE WHEN rel_outband.branch IN ["-global-", $branch] THEN [1] ELSE [] END |
            SET rel_outband.to = $current_time
        )
        WITH DISTINCT active_node, new_node
        // Process Inbound Relationship
        MATCH (active_node)<-[]-(peer)
        WITH DISTINCT active_node, new_node, peer
        CALL (active_node, peer) {
            MATCH (active_node)<-[r]-(peer)
            WHERE %(branch_filter)s
            RETURN active_node as n1, r as rel_inband1, peer as p1
            ORDER BY r.branch_level DESC, r.from DESC
            LIMIT 1
        }
        WITH n1 as active_node, rel_inband1 as rel_inband, p1 as peer_node, new_node
        WHERE rel_inband.status = "active" AND rel_inband.to IS NULL
        CALL (%(sub_query_in_args)s) {
            %(sub_query_in)s
        }
        WITH p2 as peer_node, rel_inband, active_node, new_node
        FOREACH (i in CASE WHEN rel_inband.branch IN ["-global-", $branch] THEN [1] ELSE [] END |
            SET rel_inband.to = $current_time
        )

        RETURN DISTINCT new_node
        """ % {
            "branch_filter": branch_filter,
            "labels": ":".join(self.new_node.labels),
            "sub_query_out": sub_query_out,
            "sub_query_in": sub_query_in,
            "sub_query_out_args": sub_query_out_args,
            "sub_query_in_args": sub_query_in_args,
        }
        self.add_to_query(query)



from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING, Sequence

from infrahub.core import registry
from infrahub.core.diff.payload_builder import get_display_labels_per_kind
from infrahub.core.migrations.shared import MigrationResult
from infrahub.core.schema import GenericSchema, NodeSchema, SchemaRoot, internal_schema
from infrahub.core.schema.manager import SchemaManager
from infrahub.core.validators.uniqueness.checker import UniquenessChecker
from infrahub.dependencies.registry import build_component_registry, get_component_registry
from infrahub.log import get_logger

from ..shared import InternalSchemaMigration, SchemaMigration

if TYPE_CHECKING:
    from infrahub.core.validators.uniqueness.model import NonUniqueNode
    from infrahub.database import InfrahubDatabase

log = get_logger()


async def validate_nulls_in_uniqueness_constraints(db: InfrahubDatabase) -> MigrationResult:
    """
    Validate any schema that include optional attributes in the uniqueness constraints

    An update to uniqueness constraint validation now handles NULL values as unique instead of ignoring them
    """

    default_branch = registry.get_branch_from_registry()
    build_component_registry()
    component_registry = get_component_registry()
    uniqueness_checker = await component_registry.get_component(UniquenessChecker, db=db, branch=default_branch)
    non_unique_nodes_by_kind: dict[str, list[NonUniqueNode]] = defaultdict(list)

    manager = SchemaManager()
    registry.schema = manager
    internal_schema_root = SchemaRoot(**internal_schema)
    manager.register_schema(schema=internal_schema_root)
    schema_branch = await manager.load_schema_from_db(db=db, branch=default_branch)
    manager.set_schema_branch(name=default_branch.name, schema=schema_branch)

    for schema_kind in schema_branch.node_names + schema_branch.generic_names_without_templates:
        schema = schema_branch.get(name=schema_kind, duplicate=False)
        if not isinstance(schema, NodeSchema | GenericSchema):
            continue

        uniqueness_constraint_paths = schema.get_unique_constraint_schema_attribute_paths(schema_branch=schema_branch)
        includes_optional_attr: bool = False

        for uniqueness_constraint_path in uniqueness_constraint_paths:
            for schema_attribute_path in uniqueness_constraint_path.attributes_paths:
                if schema_attribute_path.attribute_schema and schema_attribute_path.attribute_schema.optional is True:
                    includes_optional_attr = True
                    break

        if not includes_optional_attr:
            continue

        non_unique_nodes = await uniqueness_checker.check_one_schema(schema=schema)
        if non_unique_nodes:
            non_unique_nodes_by_kind[schema_kind] = non_unique_nodes

    if not non_unique_nodes_by_kind:
        return MigrationResult()

    error_strings = []
    for schema_kind, non_unique_nodes in non_unique_nodes_by_kind.items():
        display_label_map = await get_display_labels_per_kind(
            db=db, kind=schema_kind, branch_name=default_branch.name, ids=[nun.node_id for nun in non_unique_nodes]
        )
        for non_unique_node in non_unique_nodes:
            display_label = display_label_map.get(non_unique_node.node_id)
            error_str = f"{display_label or ''}({non_unique_node.node_schema.kind} / {non_unique_node.node_id})"
            error_str += " violates uniqueness constraints for the following attributes: "
            attr_values = [
                f"{attr.attribute_name}={attr.attribute_value}" for attr in non_unique_node.non_unique_attributes
            ]
            error_str += ", ".join(attr_values)
            error_strings.append(error_str)
    if error_strings:
        error_str = "For the following nodes, you must update the uniqueness_constraints on the schema of the node"
        error_str += " to remove the attribute(s) with NULL values or update the data on the nodes to be unique"
        error_str += " now that NULL values are considered during uniqueness validation"
        return MigrationResult(errors=[error_str] + error_strings)
    return MigrationResult()


class Migration018(InternalSchemaMigration):
    name: str = "018_validate_nulls_in_uniqueness_constraints"
    minimum_version: int = 17
    migrations: Sequence[SchemaMigration] = []

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:  # noqa: ARG002
        return MigrationResult()

    async def execute(self, db: InfrahubDatabase) -> MigrationResult:
        return await validate_nulls_in_uniqueness_constraints(db=db)



from infrahub.core.node.delete_validator import NodeDeleteValidator
from infrahub.dependencies.interface import DependencyBuilder, DependencyBuilderContext


class NodeDeleteValidatorDependency(DependencyBuilder[NodeDeleteValidator]):
    @classmethod
    def build(cls, context: DependencyBuilderContext) -> NodeDeleteValidator:
        return NodeDeleteValidator(db=context.db, branch=context.branch)



from __future__ import annotations

from typing import TYPE_CHECKING

from pydantic import ConfigDict, Field

from infrahub.core.constants import Severity  # noqa: TC001
from infrahub.core.node.standard import StandardNode
from infrahub.core.query.task_log import TaskLogNodeCreateQuery
from infrahub.core.timestamp import current_timestamp

if TYPE_CHECKING:
    from infrahub.core.query.standard_node import StandardNodeQuery


class TaskLog(StandardNode):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    message: str = Field(..., description="The message of the log entry.")
    severity: Severity = Field(..., description="Severity of the event")
    task_id: str = Field(..., description="The ID of the associated task")
    timestamp: str = Field(default_factory=current_timestamp, description="The time when this task was created")

    _exclude_attrs: list[str] = ["id", "uuid", "task_id", "_query"]
    _query: type[StandardNodeQuery] = TaskLogNodeCreateQuery



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Generator

from pydantic import BaseModel, ConfigDict

from infrahub.core import registry
from infrahub.core.constants import InfrahubKind, RelationshipStatus
from infrahub.core.query import Query, QueryType

if TYPE_CHECKING:
    from infrahub.core.protocols import CoreNumberPool
    from infrahub.database import InfrahubDatabase


class NumberPoolIdentifierData(BaseModel):
    model_config = ConfigDict(frozen=True)

    value: int
    identifier: str


class IPAddressPoolGetIdentifiers(Query):
    name = "ipaddresspool_get_identifiers"
    type = QueryType.READ

    def __init__(
        self,
        pool_id: str,
        allocated: list[str],
        **kwargs: dict[str, Any],
    ) -> None:
        self.pool_id = pool_id
        self.addresses = allocated

        super().__init__(**kwargs)  # type: ignore[arg-type]

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        self.params["pool_id"] = self.pool_id
        self.params["addresses"] = self.addresses

        query = """
        MATCH (pool:%(ipaddress_pool)s { uuid: $pool_id })-[reservation:IS_RESERVED]->(allocated:BuiltinIPAddress)
        WHERE allocated.uuid in $addresses
        """ % {"ipaddress_pool": InfrahubKind.IPADDRESSPOOL}
        self.add_to_query(query)
        self.return_labels = ["allocated", "reservation"]


class IPAddressPoolGetReserved(Query):
    name = "ipaddresspool_get_reserved"
    type = QueryType.READ

    def __init__(
        self,
        pool_id: str,
        identifier: str,
        **kwargs: dict[str, Any],
    ) -> None:
        self.pool_id = pool_id
        self.identifier = identifier

        super().__init__(**kwargs)  # type: ignore[arg-type]

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        self.params["pool_id"] = self.pool_id
        self.params["identifier"] = self.identifier

        query = """
        MATCH (pool:%(ipaddress_pool)s { uuid: $pool_id })-[rel:IS_RESERVED]->(address:BuiltinIPAddress)
        WHERE rel.identifier = $identifier
        """ % {"ipaddress_pool": InfrahubKind.IPADDRESSPOOL}
        self.add_to_query(query)
        self.return_labels = ["address"]


class IPAddressPoolSetReserved(Query):
    name = "ipaddresspool_set_reserved"
    type = QueryType.WRITE

    def __init__(
        self,
        pool_id: str,
        address_id: str,
        identifier: str,
        **kwargs: dict[str, Any],
    ) -> None:
        self.pool_id = pool_id
        self.address_id = address_id
        self.identifier = identifier

        super().__init__(**kwargs)  # type: ignore[arg-type]

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        self.params["pool_id"] = self.pool_id
        self.params["address_id"] = self.address_id
        self.params["identifier"] = self.identifier

        global_branch = registry.get_global_branch()
        self.params["rel_prop"] = {
            "branch": global_branch.name,
            "branch_level": global_branch.hierarchy_level,
            "status": RelationshipStatus.ACTIVE.value,
            "from": self.at.to_string(),
            "identifier": self.identifier,
        }

        query = """
        MATCH (pool:%(ipaddress_pool)s { uuid: $pool_id })
        MATCH (address:Node { uuid: $address_id })
        CREATE (pool)-[rel:IS_RESERVED $rel_prop]->(address)
        """ % {"ipaddress_pool": InfrahubKind.IPADDRESSPOOL}

        self.add_to_query(query)
        self.return_labels = ["pool", "rel", "address"]


class NumberPoolGetAllocated(Query):
    name = "numberpool_get_allocated"
    type = QueryType.READ

    def __init__(
        self,
        pool: CoreNumberPool,
        **kwargs: dict[str, Any],
    ) -> None:
        self.pool = pool

        super().__init__(**kwargs)  # type: ignore[arg-type]

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        self.params["node_attribute"] = self.pool.node_attribute.value
        self.params["start_range"] = self.pool.start_range.value
        self.params["end_range"] = self.pool.end_range.value
        self.params["pool_id"] = self.pool.get_id()

        branch_filter, branch_params = self.branch.get_query_filter_path(
            at=self.at.to_string(), branch_agnostic=self.branch_agnostic
        )
        self.params.update(branch_params)

        query = """
        MATCH (n:%(node)s)-[ha:HAS_ATTRIBUTE]-(a:Attribute {name: $node_attribute})-[hv:HAS_VALUE]-(av:AttributeValueIndexed)
        MATCH (a)-[hs:HAS_SOURCE]-(pool:%(number_pool_kind)s)
        WHERE
            pool.uuid = $pool_id
            AND av.value >= $start_range and av.value <= $end_range
            AND all(r in [ha, hv, hs] WHERE (%(branch_filter)s))
            AND ha.status = "active"
            AND hv.status = "active"
            AND hs.status = "active"
        """ % {
            "node": self.pool.node.value,
            "number_pool_kind": InfrahubKind.NUMBERPOOL,
            "branch_filter": branch_filter,
        }
        self.add_to_query(query)

        self.return_labels = ["n.uuid as id", "hv.branch as branch", "av.value as value"]
        self.order_by = ["av.value"]


class NumberPoolGetReserved(Query):
    name = "numberpool_get_reserved"
    type = QueryType.READ

    def __init__(
        self,
        pool_id: str,
        identifier: str | None = None,
        **kwargs: dict[str, Any],
    ) -> None:
        self.pool_id = pool_id
        self.identifier = identifier

        super().__init__(**kwargs)  # type: ignore[arg-type]

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        self.params["pool_id"] = self.pool_id
        self.params["identifier"] = self.identifier

        branch_filter, branch_params = self.branch.get_query_filter_path(
            at=self.at.to_string(), branch_agnostic=self.branch_agnostic
        )

        self.params.update(branch_params)

        # If identifier is not provided, we return all reservations for the pool
        identifier_filter = ""
        if self.identifier:
            identifier_filter = "r.identifier = $identifier AND "
            self.params["identifier"] = self.identifier

        query = """
        MATCH (pool:%(number_pool)s { uuid: $pool_id })-[r:IS_RESERVED]->(reservation:AttributeValue)
        WHERE
            %(identifier_filter)s
            %(branch_filter)s
        """ % {
            "branch_filter": branch_filter,
            "number_pool": InfrahubKind.NUMBERPOOL,
            "identifier_filter": identifier_filter,
        }
        self.add_to_query(query)
        self.return_labels = ["reservation.value AS value", "r.identifier AS identifier"]

    def get_reservation(self) -> int | None:
        result = self.get_result()
        if result:
            return result.get_as_optional_type("value", return_type=int)
        return None

    def get_reservations(self) -> Generator[NumberPoolIdentifierData]:
        for result in self.results:
            yield NumberPoolIdentifierData.model_construct(
                value=result.get_as_type("value", return_type=int),
                identifier=result.get_as_type("identifier", return_type=str),
            )


class PoolChangeReserved(Query):
    """Change the identifier on all pools.
    This is useful when a node is being converted to a different type and its ID has changed
    """

    name = "pool_change_reserved"
    type = QueryType.WRITE

    def __init__(
        self,
        existing_identifier: str,
        new_identifier: str,
        **kwargs: dict[str, Any],
    ) -> None:
        self.existing_identifier = existing_identifier
        self.new_identifier = new_identifier

        super().__init__(**kwargs)  # type: ignore[arg-type]

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        self.params["new_identifier"] = self.new_identifier
        self.params["existing_identifier"] = self.existing_identifier
        self.params["at"] = self.at.to_string()

        branch_filter, branch_params = self.branch.get_query_filter_path(
            at=self.at.to_string(), branch_agnostic=self.branch_agnostic
        )

        self.params.update(branch_params)

        global_branch = registry.get_global_branch()
        self.params["rel_prop"] = {
            "branch": global_branch.name,
            "branch_level": global_branch.hierarchy_level,
            "status": RelationshipStatus.ACTIVE.value,
            "from": self.at.to_string(),
            "identifier": self.new_identifier,
        }

        query = """
        MATCH (pool:Node)-[r:IS_RESERVED]->(resource)
        WHERE
            r.identifier = $existing_identifier
            AND
            %(branch_filter)s
        SET r.to = $at
        CREATE (pool)-[new_rel:IS_RESERVED $rel_prop]->(resource)
        """ % {"branch_filter": branch_filter}
        self.add_to_query(query)
        self.return_labels = ["pool.uuid AS pool_id", "r", "new_rel"]


"""
Important!: The relationship IS_RESERVED for Number is not being cleaned up when the node or the branch is deleted
I think this is something we should address in the future.
It works for now because the query has been updated to match the identifier in IS_RESERVED with the UUID of the related node
But in the future, if we need to use an identifier that is not the UUID, we will need to clean up the relationships
This will be especially important as we want to support upsert with NumberPool
"""


class NumberPoolGetUsed(Query):
    name = "number_pool_get_used"
    type = QueryType.READ
    return_model = NumberPoolIdentifierData

    def __init__(
        self,
        pool: CoreNumberPool,
        **kwargs: dict[str, Any],
    ) -> None:
        self.pool = pool

        super().__init__(**kwargs)  # type: ignore[arg-type]

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        self.params["pool_id"] = self.pool.get_id()
        self.params["start_range"] = self.pool.start_range.value
        self.params["end_range"] = self.pool.end_range.value

        branch_filter, branch_params = self.branch.get_query_filter_path(
            at=self.at.to_string(), branch_agnostic=self.branch_agnostic
        )

        self.params.update(branch_params)
        self.params["attribute_name"] = self.pool.node_attribute.value

        query = """
        MATCH (pool:%(number_pool)s { uuid: $pool_id })-[res:IS_RESERVED]->(av:AttributeValueIndexed)
        WHERE toInteger(av.value) >= $start_range and toInteger(av.value) <= $end_range
        CALL (pool, res, av) {
            MATCH (pool)-[res]->(av)<-[hv:HAS_VALUE]-(attr:Attribute)<-[ha:HAS_ATTRIBUTE]-(n:%(node)s)
            WHERE
                n.uuid = res.identifier AND
                attr.name = $attribute_name AND
                all(r in [res, hv, ha] WHERE (%(branch_filter)s))
            ORDER BY res.branch_level DESC, hv.branch_level DESC, ha.branch_level DESC, res.from DESC, hv.from DESC, ha.from DESC
            RETURN (res.status = "active" AND hv.status = "active" AND ha.status = "active") AS is_active
            LIMIT 1
        }
        WITH av, res, is_active
        WHERE is_active = True
        """ % {
            "branch_filter": branch_filter,
            "number_pool": InfrahubKind.NUMBERPOOL,
            "node": self.pool.node.value,
        }

        self.add_to_query(query)
        self.return_labels = ["DISTINCT(av.value) as value", "res.identifier as identifier"]
        self.order_by = ["value"]

    def iter_results(self) -> Generator[NumberPoolIdentifierData]:
        for result in self.results:
            yield self.return_model.model_construct(
                value=result.get_as_type("value", return_type=int),
                identifier=result.get_as_type("identifier", return_type=str),
            )


class NumberPoolSetReserved(Query):
    name = "numberpool_set_reserved"
    type = QueryType.WRITE

    def __init__(
        self,
        pool_id: str,
        reserved: int,
        identifier: str,
        **kwargs: dict[str, Any],
    ) -> None:
        self.pool_id = pool_id
        self.reserved = reserved
        self.identifier = identifier

        super().__init__(**kwargs)  # type: ignore[arg-type]

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        self.params["pool_id"] = self.pool_id
        self.params["reserved"] = self.reserved
        self.params["identifier"] = self.identifier

        global_branch = registry.get_global_branch()
        self.params["rel_prop"] = {
            "branch": global_branch.name,
            "branch_level": global_branch.hierarchy_level,
            "status": RelationshipStatus.ACTIVE.value,
            "from": self.at.to_string(),
            "identifier": self.identifier,
        }

        query = """
        MATCH (pool:%(number_pool)s { uuid: $pool_id })
        MERGE (value:AttributeValue:AttributeValueIndexed { value: $reserved, is_default: false })
        WITH value, pool
        LIMIT 1
        CREATE (pool)-[rel:IS_RESERVED $rel_prop]->(value)
        """ % {"number_pool": InfrahubKind.NUMBERPOOL}

        self.add_to_query(query)
        self.return_labels = ["value"]


class PrefixPoolGetIdentifiers(Query):
    name = "prefixpool_get_identifiers"
    type = QueryType.READ

    def __init__(
        self,
        pool_id: str,
        allocated: list[str],
        **kwargs: dict[str, Any],
    ) -> None:
        self.pool_id = pool_id
        self.prefixes = allocated

        super().__init__(**kwargs)  # type: ignore[arg-type]

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        self.params["pool_id"] = self.pool_id
        self.params["prefixes"] = self.prefixes

        query = """
        MATCH (pool:%(ipaddress_pool)s { uuid: $pool_id })-[reservation:IS_RESERVED]->(allocated:BuiltinIPPrefix)
        WHERE allocated.uuid in $prefixes
        """ % {"ipaddress_pool": InfrahubKind.IPPREFIXPOOL}
        self.add_to_query(query)
        self.return_labels = ["allocated", "reservation"]


class PrefixPoolGetReserved(Query):
    name = "prefixpool_get_reserved"
    type = QueryType.READ

    def __init__(
        self,
        pool_id: str,
        identifier: str,
        **kwargs: dict[str, Any],
    ) -> None:
        self.pool_id = pool_id
        self.identifier = identifier

        super().__init__(**kwargs)  # type: ignore[arg-type]

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        self.params["pool_id"] = self.pool_id
        self.params["identifier"] = self.identifier

        query = """
        MATCH (pool:%(prefix_pool)s { uuid: $pool_id })-[rel:IS_RESERVED]->(prefix:BuiltinIPPrefix)
        WHERE rel.identifier = $identifier
        """ % {"prefix_pool": InfrahubKind.IPPREFIXPOOL}
        self.add_to_query(query)
        self.return_labels = ["prefix"]


class PrefixPoolSetReserved(Query):
    name = "prefixpool_set_reserved"
    type = QueryType.WRITE

    def __init__(
        self,
        pool_id: str,
        prefix_id: str,
        identifier: str,
        **kwargs: dict[str, Any],
    ) -> None:
        self.pool_id = pool_id
        self.prefix_id = prefix_id
        self.identifier = identifier

        super().__init__(**kwargs)  # type: ignore[arg-type]

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        self.params["pool_id"] = self.pool_id
        self.params["prefix_id"] = self.prefix_id
        self.params["identifier"] = self.identifier

        global_branch = registry.get_global_branch()
        self.params["rel_prop"] = {
            "branch": global_branch.name,
            "branch_level": global_branch.hierarchy_level,
            "status": RelationshipStatus.ACTIVE.value,
            "from": self.at.to_string(),
            "identifier": self.identifier,
        }

        query = """
        MATCH (pool:%(prefix_pool)s { uuid: $pool_id })
        MATCH (prefix:Node { uuid: $prefix_id })
        CREATE (pool)-[rel:IS_RESERVED $rel_prop]->(prefix)
        """ % {"prefix_pool": InfrahubKind.IPPREFIXPOOL}

        self.add_to_query(query)
        self.return_labels = ["pool", "rel", "prefix"]



from __future__ import annotations

import asyncio
from collections.abc import AsyncGenerator, Awaitable, Generator
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Callable

if TYPE_CHECKING:
    from .node import InfrahubNode, InfrahubNodeSync


@dataclass
class BatchTask:
    task: Callable[[Any], Awaitable[Any]]
    args: tuple[Any, ...]
    kwargs: dict[str, Any]
    node: Any | None = None


@dataclass
class BatchTaskSync:
    task: Callable[..., Any]
    args: tuple[Any, ...]
    kwargs: dict[str, Any]
    node: InfrahubNodeSync | None = None

    def execute(self, return_exceptions: bool = False) -> tuple[InfrahubNodeSync | None, Any]:
        """Executes the stored task."""
        result = None
        try:
            result = self.task(*self.args, **self.kwargs)
        except Exception as exc:
            if return_exceptions:
                return self.node, exc
            raise exc

        return self.node, result


async def execute_batch_task_in_pool(
    task: BatchTask, semaphore: asyncio.Semaphore, return_exceptions: bool = False
) -> tuple[InfrahubNode | None, Any]:
    async with semaphore:
        try:
            result = await task.task(*task.args, **task.kwargs)
        except Exception as exc:
            if return_exceptions:
                return (task.node, exc)
            raise exc

        return (task.node, result)


class InfrahubBatch:
    def __init__(
        self,
        semaphore: asyncio.Semaphore | None = None,
        max_concurrent_execution: int = 5,
        return_exceptions: bool = False,
    ):
        self._tasks: list[BatchTask] = []
        self.semaphore = semaphore or asyncio.Semaphore(value=max_concurrent_execution)
        self.return_exceptions = return_exceptions

    @property
    def num_tasks(self) -> int:
        return len(self._tasks)

    def add(self, *args: Any, task: Callable, node: Any | None = None, **kwargs: Any) -> None:
        self._tasks.append(BatchTask(task=task, node=node, args=args, kwargs=kwargs))

    async def execute(self) -> AsyncGenerator:
        tasks = []

        for batch_task in self._tasks:
            tasks.append(
                asyncio.create_task(
                    execute_batch_task_in_pool(
                        task=batch_task, semaphore=self.semaphore, return_exceptions=self.return_exceptions
                    )
                )
            )

        for completed_task in asyncio.as_completed(tasks):
            node, result = await completed_task
            if isinstance(result, Exception) and not self.return_exceptions:
                raise result
            yield node, result


class InfrahubBatchSync:
    def __init__(self, max_concurrent_execution: int = 5, return_exceptions: bool = False):
        self._tasks: list[BatchTaskSync] = []
        self.max_concurrent_execution = max_concurrent_execution
        self.return_exceptions = return_exceptions

    @property
    def num_tasks(self) -> int:
        return len(self._tasks)

    def add(self, *args: Any, task: Callable[..., Any], node: Any | None = None, **kwargs: Any) -> None:
        self._tasks.append(BatchTaskSync(task=task, node=node, args=args, kwargs=kwargs))

    def execute(self) -> Generator[tuple[InfrahubNodeSync | None, Any], None, None]:
        with ThreadPoolExecutor(max_workers=self.max_concurrent_execution) as executor:
            futures = [executor.submit(task.execute, return_exceptions=self.return_exceptions) for task in self._tasks]
            for future in futures:
                node, result = future.result()
                if isinstance(result, Exception) and not self.return_exceptions:
                    raise result
                yield node, result



from typing import ClassVar

from pydantic import Field

from infrahub.core.constants import InfrahubKind, MutationAction

from .constants import EVENT_NAMESPACE
from .models import InfrahubEvent


class ProposedChangeEvent(InfrahubEvent):
    proposed_change_id: str = Field(..., description="The ID of the proposed change")
    proposed_change_name: str = Field(..., description="The name of the proposed change")
    proposed_change_state: str = Field(..., description="The state of the proposed change")

    def get_resource(self) -> dict[str, str]:
        return {
            "prefect.resource.id": f"infrahub.proposed_change.{self.proposed_change_id}",
            "infrahub.node.kind": InfrahubKind.PROPOSEDCHANGE,
            "infrahub.node.id": self.proposed_change_id,
            "infrahub.proposed_change.name": self.proposed_change_name,
            "infrahub.proposed_change.state": self.proposed_change_state,
            "infrahub.branch.name": self.meta.context.branch.name,
        }


class ProposedChangeReviewEvent(ProposedChangeEvent):
    reviewer_account_id: str = Field(..., description="The ID of the user who reviewed the proposed change")
    reviewer_account_name: str = Field(..., description="The name of the user who reviewed the proposed change")
    reviewer_decision: str = Field(..., description="The decision made by the reviewer")

    def get_related(self) -> list[dict[str, str]]:
        related = super().get_related()
        related.append(
            {
                "prefect.resource.id": self.reviewer_account_id,
                "prefect.resource.role": "infrahub.related.node",
                "infrahub.node.kind": InfrahubKind.GENERICACCOUNT,
                "infrahub.node.id": self.reviewer_account_id,
                "infrahub.reviewer.account.name": self.reviewer_account_name,
            }
        )
        return related

    def get_resource(self) -> dict[str, str]:
        return {**super().get_resource(), "infrahub.proposed_change.reviewer_decision": self.reviewer_decision}


class ProposedChangeReviewRevokedEvent(ProposedChangeEvent):
    reviewer_account_id: str = Field(..., description="The ID of the user who reviewed the proposed change")
    reviewer_account_name: str = Field(..., description="The name of the user who reviewed the proposed change")
    reviewer_former_decision: str = Field(..., description="The former decision made by the reviewer")

    def get_related(self) -> list[dict[str, str]]:
        related = super().get_related()
        related.append(
            {
                "prefect.resource.id": self.reviewer_account_id,
                "prefect.resource.role": "infrahub.related.node",
                "infrahub.node.kind": InfrahubKind.GENERICACCOUNT,
                "infrahub.node.id": self.reviewer_account_id,
                "infrahub.reviewer.account.name": self.reviewer_account_name,
            }
        )
        return related

    def get_resource(self) -> dict[str, str]:
        return {
            **super().get_resource(),
            "infrahub.proposed_change.reviewer_former_decision": self.reviewer_former_decision,
        }


class ProposedChangeMergedEvent(ProposedChangeEvent):
    """Event generated when a proposed change has been merged"""

    event_name: ClassVar[str] = f"{EVENT_NAMESPACE}.proposed_change.merged"

    merged_by_account_id: str = Field(..., description="The ID of the user who merged the proposed change")
    merged_by_account_name: str = Field(..., description="The name of the user who merged the proposed change")

    def get_related(self) -> list[dict[str, str]]:
        related = super().get_related()
        related.append(
            {
                "prefect.resource.id": self.merged_by_account_id,
                "prefect.resource.role": "infrahub.related.node",
                "infrahub.node.kind": InfrahubKind.GENERICACCOUNT,
                "infrahub.node.id": self.merged_by_account_id,
                "infrahub.merged_by.account.name": self.merged_by_account_name,
            }
        )
        return related

    def get_resource(self) -> dict[str, str]:
        return {
            **super().get_resource(),
            "infrahub.proposed_change.merged_by_account_id": self.merged_by_account_id,
            "infrahub.proposed_change.merged_by_account_name": self.merged_by_account_name,
        }


class ProposedChangeReviewRequestedEvent(ProposedChangeEvent):
    """Event generated when a proposed change has been flagged for review"""

    event_name: ClassVar[str] = f"{EVENT_NAMESPACE}.proposed_change.review_requested"

    requested_by_account_id: str = Field(
        ..., description="The ID of the user who requested the proposed change to be reviewed"
    )
    requested_by_account_name: str = Field(
        ..., description="The name of the user who requested the proposed change to be reviewed"
    )

    def get_related(self) -> list[dict[str, str]]:
        related = super().get_related()
        related.append(
            {
                "prefect.resource.id": self.requested_by_account_id,
                "prefect.resource.role": "infrahub.related.node",
                "infrahub.node.kind": InfrahubKind.GENERICACCOUNT,
                "infrahub.node.id": self.requested_by_account_id,
                "infrahub.requested_by.account.name": self.requested_by_account_name,
            }
        )
        return related


class ProposedChangeApprovedEvent(ProposedChangeReviewEvent):
    """Event generated when a proposed change has been approved"""

    event_name: ClassVar[str] = f"{EVENT_NAMESPACE}.proposed_change.approved"


class ProposedChangeRejectedEvent(ProposedChangeReviewEvent):
    """Event generated when a proposed change has been rejected"""

    event_name: ClassVar[str] = f"{EVENT_NAMESPACE}.proposed_change.rejected"


class ProposedChangeApprovalRevokedEvent(ProposedChangeReviewRevokedEvent):
    """Event generated when a proposed change approval has been revoked"""

    event_name: ClassVar[str] = f"{EVENT_NAMESPACE}.proposed_change.approval_revoked"


class ProposedChangeRejectionRevokedEvent(ProposedChangeReviewRevokedEvent):
    """Event generated when a proposed change rejection has been revoked"""

    event_name: ClassVar[str] = f"{EVENT_NAMESPACE}.proposed_change.rejection_revoked"


class ProposedChangeThreadEvent(ProposedChangeEvent):
    thread_id: str = Field(..., description="The ID of the thread that was created or updated")
    thread_kind: str = Field(..., description="The name of the thread that was created or updated")

    def get_related(self) -> list[dict[str, str]]:
        related = super().get_related()
        related.append(
            {
                "prefect.resource.id": self.thread_id,
                "prefect.resource.role": "infrahub.related.node",
                "infrahub.node.kind": self.thread_kind,
                "infrahub.node.id": self.thread_id,
            }
        )
        return related


class ProposedChangeThreadCreatedEvent(ProposedChangeThreadEvent):
    """Event generated when a thread has been created in a proposed change"""

    event_name: ClassVar[str] = f"{EVENT_NAMESPACE}.proposed_change_thread.created"
    action: MutationAction = MutationAction.CREATED


class ProposedChangeThreadUpdatedEvent(ProposedChangeThreadEvent):
    """Event generated when a thread has been updated in a proposed change"""

    event_name: ClassVar[str] = f"{EVENT_NAMESPACE}.proposed_change_thread.updated"
    action: MutationAction = MutationAction.UPDATED



from typing import Any

from prefect import flow

from infrahub.git.repository import get_initialized_repo
from infrahub.log import get_logger
from infrahub.workers.dependencies import get_client
from infrahub.workflows.utils import add_branch_tag

from .models import TransformJinjaTemplateData, TransformPythonData

log = get_logger()


@flow(name="transform_render_python", flow_run_name="Render transform python", persist_result=True)
async def transform_python(message: TransformPythonData) -> Any:
    await add_branch_tag(branch_name=message.branch)

    client = get_client()

    repo = await get_initialized_repo(
        client=client,
        repository_id=message.repository_id,
        name=message.repository_name,
        repository_kind=message.repository_kind,
        commit=message.commit,
    )

    transformed_data = await repo.execute_python_transform.with_options(timeout_seconds=message.timeout)(
        client=client,
        branch_name=message.branch,
        commit=message.commit,
        location=message.transform_location,
        data=message.data,
        convert_query_response=message.convert_query_response,
    )  # type: ignore[misc]

    return transformed_data


@flow(name="transform_render_jinja2_template", flow_run_name="Render transform Jinja2", persist_result=True)
async def transform_render_jinja2_template(message: TransformJinjaTemplateData) -> str:
    await add_branch_tag(branch_name=message.branch)

    client = get_client()

    repo = await get_initialized_repo(
        client=client,
        repository_id=message.repository_id,
        name=message.repository_name,
        repository_kind=message.repository_kind,
        commit=message.commit,
    )

    rendered_template = await repo.render_jinja2_template.with_options(timeout_seconds=message.timeout)(
        commit=message.commit, location=message.template_location, data={"data": message.data}
    )  # type: ignore[misc]

    return rendered_template



from __future__ import annotations

from .attribute import Attribute
from .constants import (
    ARTIFACT_DEFINITION_GENERATE_FEATURE_NOT_SUPPORTED_MESSAGE,
    ARTIFACT_FETCH_FEATURE_NOT_SUPPORTED_MESSAGE,
    ARTIFACT_GENERATE_FEATURE_NOT_SUPPORTED_MESSAGE,
    HFID_STR_SEPARATOR,
    IP_TYPES,
    PROPERTIES_FLAG,
    PROPERTIES_OBJECT,
    SAFE_VALUE,
)
from .node import InfrahubNode, InfrahubNodeBase, InfrahubNodeSync
from .parsers import parse_human_friendly_id
from .property import NodeProperty
from .related_node import RelatedNode, RelatedNodeBase, RelatedNodeSync
from .relationship import RelationshipManager, RelationshipManagerBase, RelationshipManagerSync

__all__ = [
    "ARTIFACT_DEFINITION_GENERATE_FEATURE_NOT_SUPPORTED_MESSAGE",
    "ARTIFACT_FETCH_FEATURE_NOT_SUPPORTED_MESSAGE",
    "ARTIFACT_GENERATE_FEATURE_NOT_SUPPORTED_MESSAGE",
    "HFID_STR_SEPARATOR",
    "IP_TYPES",
    "PROPERTIES_FLAG",
    "PROPERTIES_OBJECT",
    "SAFE_VALUE",
    "Attribute",
    "InfrahubNode",
    "InfrahubNodeBase",
    "InfrahubNodeSync",
    "NodeProperty",
    "RelatedNode",
    "RelatedNodeBase",
    "RelatedNodeSync",
    "RelationshipManager",
    "RelationshipManagerBase",
    "RelationshipManagerSync",
    "parse_human_friendly_id",
]



from infrahub.core.constants import (
    AllowOverrideType,
    BranchSupportType,
    EventType,
    InfrahubKind,
)
from infrahub.core.constants import RelationshipCardinality as Cardinality
from infrahub.core.constants import RelationshipKind as RelKind

from ...attribute_schema import AttributeSchema as Attr
from ...dropdown import DropdownChoice
from ...generic_schema import GenericSchema
from ...node_schema import NodeSchema
from ...relationship_schema import (
    RelationshipSchema as Rel,
)

core_webhook = GenericSchema(
    name="Webhook",
    namespace="Core",
    description="A webhook that connects to an external integration",
    label="Webhook",
    default_filter="name__value",
    order_by=["name__value"],
    display_labels=["name__value"],
    include_in_menu=False,
    icon="mdi:webhook",
    branch=BranchSupportType.AGNOSTIC,
    uniqueness_constraints=[["name__value"]],
    attributes=[
        Attr(name="name", kind="Text", unique=True, order_weight=1000),
        Attr(
            name="event_type",
            kind="Text",
            enum=["all"] + EventType.available_types(),
            default_value="all",
            order_weight=1500,
            description="The event type that triggers the webhook",
        ),
        Attr(
            name="branch_scope",
            kind="Dropdown",
            choices=[
                DropdownChoice(
                    name="all_branches",
                    label="All Branches",
                    description="All branches",
                    color="#fef08a",
                ),
                DropdownChoice(
                    name="default_branch",
                    label="Default Branch",
                    description="Only the default branch",
                    color="#86efac",
                ),
                DropdownChoice(
                    name="other_branches",
                    label="Other Branches",
                    description="All branches except the default branch",
                    color="#e5e7eb",
                ),
            ],
            default_value="default_branch",
            optional=False,
            order_weight=2000,
            allow_override=AllowOverrideType.NONE,
        ),
        Attr(
            name="node_kind",
            kind="Text",
            optional=True,
            description="Only send node mutation events for nodes of this kind",
            order_weight=2250,
        ),
        Attr(
            name="description",
            kind="Text",
            optional=True,
            order_weight=2500,
        ),
        Attr(name="url", kind="URL", order_weight=3000),
        Attr(
            name="validate_certificates",
            kind="Boolean",
            default_value=True,
            optional=True,
            order_weight=5000,
        ),
    ],
)

core_standard_webhook = NodeSchema(
    name="StandardWebhook",
    namespace="Core",
    description="A webhook that connects to an external integration",
    label="Standard Webhook",
    default_filter="name__value",
    order_by=["name__value"],
    display_labels=["name__value"],
    include_in_menu=False,
    icon="mdi:webhook",
    branch=BranchSupportType.AGNOSTIC,
    generate_profile=False,
    inherit_from=[InfrahubKind.WEBHOOK, InfrahubKind.TASKTARGET],
    attributes=[
        Attr(name="shared_key", kind="Password", unique=False, order_weight=4000),
    ],
)

core_custom_webhook = NodeSchema(
    name="CustomWebhook",
    namespace="Core",
    description="A webhook that connects to an external integration",
    label="Custom Webhook",
    default_filter="name__value",
    order_by=["name__value"],
    display_labels=["name__value"],
    include_in_menu=False,
    icon="mdi:cog-outline",
    branch=BranchSupportType.AGNOSTIC,
    generate_profile=False,
    inherit_from=[InfrahubKind.WEBHOOK, InfrahubKind.TASKTARGET],
    attributes=[
        Attr(name="shared_key", kind="Password", unique=False, optional=True, order_weight=4000),
    ],
    relationships=[
        Rel(
            name="transformation",
            peer=InfrahubKind.TRANSFORMPYTHON,
            kind=RelKind.ATTRIBUTE,
            identifier="webhook___transformation",
            cardinality=Cardinality.ONE,
            optional=True,
            order_weight=7000,
        ),
    ],
)



from pathlib import Path

from .edge_adder import PatchPlanEdgeAdder
from .edge_deleter import PatchPlanEdgeDeleter
from .edge_updater import PatchPlanEdgeUpdater
from .models import EdgeToAdd, EdgeToDelete, EdgeToUpdate, PatchPlan, VertexToAdd, VertexToDelete, VertexToUpdate
from .plan_reader import PatchPlanReader
from .plan_writer import PatchPlanWriter
from .queries.base import PatchQuery
from .vertex_adder import PatchPlanVertexAdder
from .vertex_deleter import PatchPlanVertexDeleter
from .vertex_updater import PatchPlanVertexUpdater


class PatchPlanEdgeDbIdTranslator:
    def translate_to_db_ids(self, patch_plan: PatchPlan) -> None:
        for edge_to_add in patch_plan.edges_to_add:
            translated_from_id = patch_plan.get_database_id_for_added_element(abstract_id=edge_to_add.from_id)
            edge_to_add.from_id = translated_from_id
            translated_to_id = patch_plan.get_database_id_for_added_element(abstract_id=edge_to_add.to_id)
            edge_to_add.to_id = translated_to_id


class PatchRunner:
    def __init__(
        self,
        plan_writer: PatchPlanWriter,
        plan_reader: PatchPlanReader,
        edge_db_id_translator: PatchPlanEdgeDbIdTranslator,
        vertex_adder: PatchPlanVertexAdder,
        vertex_updater: PatchPlanVertexUpdater,
        vertex_deleter: PatchPlanVertexDeleter,
        edge_adder: PatchPlanEdgeAdder,
        edge_updater: PatchPlanEdgeUpdater,
        edge_deleter: PatchPlanEdgeDeleter,
    ) -> None:
        self.plan_writer = plan_writer
        self.plan_reader = plan_reader
        self.edge_db_id_translator = edge_db_id_translator
        self.vertex_adder = vertex_adder
        self.vertex_updater = vertex_updater
        self.vertex_deleter = vertex_deleter
        self.edge_adder = edge_adder
        self.edge_updater = edge_updater
        self.edge_deleter = edge_deleter

    async def prepare_plan(self, patch_query: PatchQuery, directory: Path) -> Path:
        patch_plan = await patch_query.plan()
        return self.plan_writer.write(patches_directory=directory, patch_plan=patch_plan)

    async def apply(self, patch_plan_directory: Path) -> PatchPlan:
        patch_plan = self.plan_reader.read(patch_plan_directory)
        await self._apply_vertices_to_add(patch_plan=patch_plan, patch_plan_directory=patch_plan_directory)
        await self._apply_edges_to_add(patch_plan=patch_plan, patch_plan_directory=patch_plan_directory)
        if patch_plan.vertices_to_update:
            await self.vertex_updater.execute(vertices_to_update=patch_plan.vertices_to_update)
        await self._apply_edges_to_delete(patch_plan=patch_plan, patch_plan_directory=patch_plan_directory)
        await self._apply_vertices_to_delete(patch_plan=patch_plan, patch_plan_directory=patch_plan_directory)
        if patch_plan.edges_to_update:
            await self.edge_updater.execute(edges_to_update=patch_plan.edges_to_update)
        return patch_plan

    async def _apply_vertices_to_add(self, patch_plan: PatchPlan, patch_plan_directory: Path) -> None:
        if not patch_plan.vertices_to_add:
            return
        unadded_vertices = [
            v for v in patch_plan.vertices_to_add if not patch_plan.has_element_been_added(v.identifier)
        ]
        try:
            async for added_element_id_map in self.vertex_adder.execute(vertices_to_add=unadded_vertices):
                patch_plan.added_element_db_id_map.update(added_element_id_map)
        finally:
            # record the added elements so that we do not double-add them if the patch is run again
            self.plan_writer.write_added_db_id_map(
                patch_plan_directory=patch_plan_directory, db_id_map=patch_plan.added_element_db_id_map
            )

    async def _apply_edges_to_add(self, patch_plan: PatchPlan, patch_plan_directory: Path) -> None:
        if not patch_plan.edges_to_add:
            return
        self.edge_db_id_translator.translate_to_db_ids(patch_plan=patch_plan)
        unadded_edges = [e for e in patch_plan.edges_to_add if not patch_plan.has_element_been_added(e.identifier)]
        try:
            async for added_element_id_map in self.edge_adder.execute(edges_to_add=unadded_edges):
                patch_plan.added_element_db_id_map.update(added_element_id_map)
        finally:
            # record the added elements so that we do not double-add them if the patch is run again
            self.plan_writer.write_added_db_id_map(
                patch_plan_directory=patch_plan_directory, db_id_map=patch_plan.added_element_db_id_map
            )

    async def _apply_vertices_to_delete(self, patch_plan: PatchPlan, patch_plan_directory: Path) -> None:
        if not patch_plan.vertices_to_delete:
            return
        try:
            async for deleted_ids in self.vertex_deleter.execute(vertices_to_delete=patch_plan.vertices_to_delete):
                patch_plan.deleted_db_ids |= deleted_ids
        finally:
            # record the deleted elements so that we know what to add if the patch is reverted
            self.plan_writer.write_deleted_db_ids(
                patch_plan_directory=patch_plan_directory, deleted_ids=patch_plan.deleted_db_ids
            )

    async def _apply_edges_to_delete(self, patch_plan: PatchPlan, patch_plan_directory: Path) -> None:
        if not patch_plan.edges_to_delete:
            return
        try:
            async for deleted_ids in self.edge_deleter.execute(edges_to_delete=patch_plan.edges_to_delete):
                patch_plan.deleted_db_ids |= deleted_ids
        finally:
            # record the deleted elements so that we know what to add if the patch is reverted
            self.plan_writer.write_deleted_db_ids(
                patch_plan_directory=patch_plan_directory, deleted_ids=patch_plan.deleted_db_ids
            )

    async def revert(self, patch_plan_directory: Path) -> PatchPlan:
        """Invert the PatchPlan to create the complement of every added/updated/deleted element and undo them"""
        patch_plan = self.plan_reader.read(patch_plan_directory)
        await self._revert_deleted_vertices(patch_plan=patch_plan, patch_plan_directory=patch_plan_directory)
        await self._revert_deleted_edges(
            patch_plan=patch_plan,
            patch_plan_directory=patch_plan_directory,
        )
        await self._revert_added_edges(patch_plan=patch_plan, patch_plan_directory=patch_plan_directory)
        await self._revert_added_vertices(patch_plan=patch_plan, patch_plan_directory=patch_plan_directory)
        vertices_to_update = [
            VertexToUpdate(
                db_id=vertex_update_to_revert.db_id,
                before_props=vertex_update_to_revert.after_props,
                after_props=vertex_update_to_revert.before_props,
            )
            for vertex_update_to_revert in patch_plan.vertices_to_update
        ]
        if vertices_to_update:
            await self.vertex_updater.execute(vertices_to_update=vertices_to_update)

        edges_to_update = [
            EdgeToUpdate(
                db_id=edge_update_to_revert.db_id,
                before_props=edge_update_to_revert.after_props,
                after_props=edge_update_to_revert.before_props,
            )
            for edge_update_to_revert in patch_plan.edges_to_update
        ]
        if edges_to_update:
            await self.edge_updater.execute(edges_to_update=edges_to_update)
        if patch_plan.reverted_deleted_db_id_map:
            patch_plan.reverted_deleted_db_id_map = {}
            self.plan_writer.write_reverted_deleted_db_id_map(
                patch_plan_directory=patch_plan_directory, db_id_map=patch_plan.reverted_deleted_db_id_map
            )
        return patch_plan

    async def _revert_added_vertices(self, patch_plan: PatchPlan, patch_plan_directory: Path) -> None:
        vertices_to_delete = [
            VertexToDelete(
                db_id=patch_plan.get_database_id_for_added_element(abstract_id=vertex_add_to_revert.identifier),
                labels=vertex_add_to_revert.labels,
                before_props=vertex_add_to_revert.after_props,
            )
            for vertex_add_to_revert in patch_plan.added_vertices
        ]
        if not vertices_to_delete:
            return
        all_deleted_ids: set[str] = set()
        try:
            async for deleted_ids in self.vertex_deleter.execute(vertices_to_delete=vertices_to_delete):
                all_deleted_ids |= deleted_ids
        finally:
            if all_deleted_ids:
                patch_plan.drop_added_db_ids(db_ids_to_drop=all_deleted_ids)
                self.plan_writer.write_added_db_id_map(
                    patch_plan_directory=patch_plan_directory, db_id_map=patch_plan.added_element_db_id_map
                )

    async def _revert_deleted_vertices(self, patch_plan: PatchPlan, patch_plan_directory: Path) -> None:
        vertices_to_add = [
            VertexToAdd(
                labels=vertex_delete_to_revert.labels,
                after_props=vertex_delete_to_revert.before_props,
                identifier=vertex_delete_to_revert.db_id,
            )
            for vertex_delete_to_revert in patch_plan.deleted_vertices
        ]
        if not vertices_to_add:
            return

        deleted_to_undeleted_db_id_map: dict[str, str] = {}
        try:
            async for added_db_id_map in self.vertex_adder.execute(vertices_to_add=vertices_to_add):
                deleted_to_undeleted_db_id_map.update(added_db_id_map)
        finally:
            if deleted_to_undeleted_db_id_map:
                patch_plan.drop_deleted_db_ids(db_ids_to_drop=set(deleted_to_undeleted_db_id_map.keys()))
                self.plan_writer.write_deleted_db_ids(
                    patch_plan_directory=patch_plan_directory, deleted_ids=patch_plan.deleted_db_ids
                )
                patch_plan.reverted_deleted_db_id_map.update(deleted_to_undeleted_db_id_map)
                self.plan_writer.write_reverted_deleted_db_id_map(
                    patch_plan_directory=patch_plan_directory, db_id_map=patch_plan.reverted_deleted_db_id_map
                )

    async def _revert_added_edges(self, patch_plan: PatchPlan, patch_plan_directory: Path) -> None:
        edges_to_delete = [
            EdgeToDelete(
                db_id=patch_plan.get_database_id_for_added_element(abstract_id=edge_add_to_revert.identifier),
                from_id=edge_add_to_revert.from_id,
                to_id=edge_add_to_revert.to_id,
                edge_type=edge_add_to_revert.edge_type,
                before_props=edge_add_to_revert.after_props,
            )
            for edge_add_to_revert in patch_plan.added_edges
        ]
        if not edges_to_delete:
            return
        all_deleted_ids: set[str] = set()
        try:
            async for deleted_ids in self.edge_deleter.execute(edges_to_delete=edges_to_delete):
                all_deleted_ids |= deleted_ids
        finally:
            if all_deleted_ids:
                patch_plan.drop_added_db_ids(db_ids_to_drop=all_deleted_ids)
                self.plan_writer.write_added_db_id_map(
                    patch_plan_directory=patch_plan_directory, db_id_map=patch_plan.added_element_db_id_map
                )

    async def _revert_deleted_edges(self, patch_plan: PatchPlan, patch_plan_directory: Path) -> None:
        edges_to_add = [
            EdgeToAdd(
                identifier=edge_delete_to_revert.db_id,
                from_id=patch_plan.reverted_deleted_db_id_map.get(
                    edge_delete_to_revert.from_id, edge_delete_to_revert.from_id
                ),
                to_id=patch_plan.reverted_deleted_db_id_map.get(
                    edge_delete_to_revert.to_id, edge_delete_to_revert.to_id
                ),
                edge_type=edge_delete_to_revert.edge_type,
                after_props=edge_delete_to_revert.before_props,
            )
            for edge_delete_to_revert in patch_plan.deleted_edges
        ]
        if not edges_to_add:
            return

        undeleted_ids: set[str] = set()
        try:
            async for added_db_id_map in self.edge_adder.execute(edges_to_add=edges_to_add):
                undeleted_ids |= set(added_db_id_map.keys())
        finally:
            if undeleted_ids:
                patch_plan.drop_deleted_db_ids(db_ids_to_drop=undeleted_ids)
                self.plan_writer.write_deleted_db_ids(
                    patch_plan_directory=patch_plan_directory, deleted_ids=patch_plan.deleted_db_ids
                )



from pydantic import BaseModel, Field


class TransformPythonData(BaseModel):
    """Sent to run a Python transform."""

    repository_id: str = Field(..., description="The unique ID of the Repository")
    repository_name: str = Field(..., description="The name of the repository")
    repository_kind: str = Field(..., description="The kind of the repository")
    data: dict = Field(..., description="Input data for the template")
    branch: str = Field(..., description="The branch to target")
    transform_location: str = Field(..., description="Location of the transform within the repository")
    commit: str = Field(..., description="The commit id to use when generating the artifact")
    convert_query_response: bool = Field(
        ..., description="Define if the GraphQL query respose should be converted into InfrahubNode objects"
    )
    timeout: int = Field(..., description="The timeout value to use when generating the artifact")


class TransformJinjaTemplateData(BaseModel):
    """Sent to trigger the checks for a repository to be executed."""

    repository_id: str = Field(..., description="The unique ID of the Repository")
    repository_name: str = Field(..., description="The name of the repository")
    repository_kind: str = Field(..., description="The kind of the repository")
    data: dict = Field(..., description="Input data for the template")
    branch: str = Field(..., description="The branch to target")
    template_location: str = Field(..., description="Location of the template within the repository")
    commit: str = Field(..., description="The commit id to use when rendering the template")
    timeout: int = Field(..., description="The timeout value to use when rendering the template")



from __future__ import annotations

from typing import Any

from graphene import BigInt, Boolean, DateTime, Field, InputObjectType, Int, List, ObjectType, String
from graphene.types.generic import GenericScalar

from infrahub.core import registry

from .enums import BranchRelativePermissionDecision
from .interface import InfrahubInterface


class GenericPoolInput(InputObjectType):
    id = String(required=True)
    identifier = String(required=False)
    data = GenericScalar(required=False)


class RelatedNodeInput(InputObjectType):
    id = String(required=False)
    hfid = Field(List(of_type=String), required=False)
    kind = String(required=False)  # Only used to resolve hfid of a related node on a generic relationship, see #4649
    from_pool = Field(GenericPoolInput, required=False)
    _relation__is_visible = Boolean(required=False)
    _relation__is_protected = Boolean(required=False)
    _relation__owner = String(required=False)
    _relation__source = String(required=False)


class IPAddressPoolInput(GenericPoolInput):
    prefixlen = Int(required=False)


class IPPrefixPoolInput(GenericPoolInput):
    size = Int(required=False)
    member_type = String(required=False)
    prefix_type = String(required=False)


class RelatedIPAddressNodeInput(InputObjectType):
    id = String(required=False)
    from_pool = Field(IPAddressPoolInput, required=False)
    _relation__is_visible = Boolean(required=False)
    _relation__is_protected = Boolean(required=False)
    _relation__owner = String(required=False)
    _relation__source = String(required=False)


class RelatedIPPrefixNodeInput(InputObjectType):
    id = String(required=False)
    hfid = Field(List(of_type=String), required=False)
    from_pool = Field(IPPrefixPoolInput, required=False)
    _relation__is_visible = Boolean(required=False)
    _relation__is_protected = Boolean(required=False)
    _relation__owner = String(required=False)
    _relation__source = String(required=False)


class PermissionType(ObjectType):
    update_value = Field(BranchRelativePermissionDecision, required=False)


class AttributeInterface(InfrahubInterface):
    is_default = Field(Boolean)
    is_inherited = Field(Boolean)
    is_protected = Field(Boolean)
    is_visible = Field(Boolean)
    updated_at = Field(DateTime)
    # Since source and owner are using a Type that is generated dynamically
    # these 2 fields will be dynamically inserted when we generate the GraphQL Schema
    # source = Field("DataSource")
    # owner = Field("DataOwner")


class BaseAttribute(ObjectType):
    id = Field(String)
    is_from_profile = Field(Boolean)
    permissions = Field(PermissionType, required=False)

    @classmethod
    def __init_subclass__(cls, **kwargs: dict[str, Any]) -> None:
        super().__init_subclass__(**kwargs)
        registry.default_graphql_type[cls.__name__] = cls


class TextAttributeType(BaseAttribute):
    value = Field(String)

    class Meta:
        description = "Attribute of type Text"
        name = "TextAttribute"
        interfaces = {AttributeInterface}


class DropdownFields(ObjectType):
    value = Field(String)
    label = Field(String)
    color = Field(String)
    description = Field(String)


class DropdownType(BaseAttribute, DropdownFields):
    class Meta:
        description = "Attribute of type Dropdown"
        name = "Dropdown"
        interfaces = {AttributeInterface}


class IPHostType(BaseAttribute):
    value = Field(String)
    ip = Field(String)
    hostmask = Field(String)
    netmask = Field(String)
    prefixlen = Field(Int)
    version = Field(Int)
    with_hostmask = Field(String)
    with_netmask = Field(String)

    class Meta:
        description = "Attribute of type IPHost"
        name = "IPHost"
        interfaces = {AttributeInterface}


class IPNetworkType(BaseAttribute):
    value = Field(String)
    broadcast_address = Field(String)
    hostmask = Field(String)
    netmask = Field(String)
    prefixlen = Field(Int)
    num_addresses = Field(Int)
    version = Field(Int)
    with_hostmask = Field(String)
    with_netmask = Field(String)

    class Meta:
        description = "Attribute of type IPNetwork"
        name = "IPNetwork"
        interfaces = {AttributeInterface}


class MacAddressType(BaseAttribute):
    value = Field(String)
    oui = Field(String)
    ei = Field(String)
    version = Field(Int)
    binary = Field(String)
    eui48 = Field(String)
    eui64 = Field(String)
    bare = Field(String, description="Format without delimiters")
    dot_notation = Field(String, description="Format often used by Cisco devices")
    semicolon_notation = Field(String, description="Format used by UNIX based systems")
    split_notation = Field(String, description="Format used by PostgreSQL")

    class Meta:
        description = "Attribute of type MacAddress"
        name = "MacAddress"
        interfaces = {AttributeInterface}


class NumberAttributeType(BaseAttribute):
    value = Field(BigInt)

    class Meta:
        description = "Attribute of type Number"
        name = "NumberAttribute"
        interfaces = {AttributeInterface}


class CheckboxAttributeType(BaseAttribute):
    value = Field(Boolean)

    class Meta:
        description = "Attribute of type Checkbox"
        name = "CheckboxAttribute"
        interfaces = {AttributeInterface}


class StrAttributeType(BaseAttribute):
    value = Field(String)

    class Meta:
        description = "Attribute of type String"
        name = "StrAttribute"
        interfaces = {AttributeInterface}


class IntAttributeType(BaseAttribute):
    value = Field(Int)

    class Meta:
        description = "Attribute of type Integer"
        name = "IntAttribute"
        interfaces = {AttributeInterface}


class BoolAttributeType(BaseAttribute):
    value = Field(Boolean)

    class Meta:
        description = "Attribute of type Boolean"
        name = "BoolAttribute"
        interfaces = {AttributeInterface}


class ListAttributeType(BaseAttribute):
    value = Field(GenericScalar)

    class Meta:
        description = "Attribute of type List"
        name = "ListAttribute"
        interfaces = {AttributeInterface}


class JSONAttributeType(BaseAttribute):
    value = Field(GenericScalar)

    class Meta:
        description = "Attribute of type JSON"
        name = "JSONAttribute"
        interfaces = {AttributeInterface}


class AnyAttributeType(BaseAttribute):
    value = Field(GenericScalar)

    class Meta:
        description = "Attribute of type GenericScalar"
        name = "AnyAttribute"
        interfaces = {AttributeInterface}



from .account import AccountPermissions, AccountToken
from .branch import BranchQueryList
from .internal import InfrahubInfo
from .ipam import (
    DeprecatedIPAddressGetNextAvailable,
    DeprecatedIPPrefixGetNextAvailable,
    InfrahubIPAddressGetNextAvailable,
    InfrahubIPPrefixGetNextAvailable,
)
from .proposed_change import ProposedChangeAvailableActions
from .relationship import Relationship
from .resource_manager import InfrahubResourcePoolAllocated, InfrahubResourcePoolUtilization
from .search import InfrahubSearchAnywhere
from .status import InfrahubStatus
from .task import Task

__all__ = [
    "AccountPermissions",
    "AccountToken",
    "BranchQueryList",
    "DeprecatedIPAddressGetNextAvailable",
    "DeprecatedIPPrefixGetNextAvailable",
    "InfrahubIPAddressGetNextAvailable",
    "InfrahubIPPrefixGetNextAvailable",
    "InfrahubInfo",
    "InfrahubResourcePoolAllocated",
    "InfrahubResourcePoolUtilization",
    "InfrahubSearchAnywhere",
    "InfrahubStatus",
    "ProposedChangeAvailableActions",
    "Relationship",
    "Task",
]
